#   include	"tedConfig.h"

#   include	<stdlib.h>

#   include	"docDraw.h"
#   include	"docLayout.h"

#   include	<appDebugon.h>

/************************************************************************/
/*									*/
/*  For dividing table rows over pages.					*/
/*									*/
/************************************************************************/

typedef struct CellDrawingProgress
    {
    int			cdpColumnAdvanced;
    int			cdpAdvanceColumnThisPage;
    int			cdpPara;
    int			cdpLinesDone;

    int			cdpX0Twips;
    int			cdpX1Twips;

    ParagraphFrame	cdpParagraphFrame;
    } CellDrawingProgress;

/************************************************************************/
/*									*/
/*  Initialise a DrawingContext						*/
/*									*/
/************************************************************************/

void docInitDrawingContext(	DrawingContext *	dc )
    {
    dc->dcCurrentPhysicalFont= -1;

    dc->dcCurrentColorSet= 0;
    dc->dcCurrentColor.rgb8Red= 0;
    dc->dcCurrentColor.rgb8Green= 0;
    dc->dcCurrentColor.rgb8Blue= 0;
    dc->dcCurrentColor.rgb8Alpha= 0;

    dc->dcDrawingData= (AppDrawingData *)0;
    dc->dcDocument= (BufferDocument *)0;
    dc->dcClipRect= (DocumentRectangle *)0;
    dc->dcDocumentSelection= (DocumentSelection *)0;
    dc->dcSelectionGeometry= (SelectionGeometry *)0;
    dc->dcFirstPage= -1;
    dc->dcLastPage= -1;
    dc->dcDrawHeadersFooters= 0;

    dc->dcDrawTableGrid= 0;

    dc->dcDrawTextLine= (DRAW_TEXT_LINE)0;
    dc->dcDrawParaTop= (DRAW_PARA_TOP)0;
    dc->dcDrawParaBottom= (DRAW_PARA_BOTTOM)0;
    dc->dcDrawParaShade= (DRAW_PARA_SHADE)0;
    dc->dcDrawCellTop= (DRAW_CELL_TOP)0;
    dc->dcDrawCellBottom= (DRAW_CELL_BOTTOM)0;
    dc->dcDrawCellLeft= (DRAW_CELL_LEFT)0;
    dc->dcDrawCellRight= (DRAW_CELL_RIGHT)0;
    dc->dcDrawCellShade= (DRAW_CELL_SHADE)0;
    dc->dcFinishPage= (FINISH_PAGE)0;
    dc->dcStartPage= (START_PAGE)0;
    dc->dcParaFramePixels= (PARA_SIDES_PIXELS)0;

    dc->dcLayoutExternal= (LAYOUT_EXTERNAL)0;
    dc->dcCloseObject= (DOC_CLOSE_OBJECT)0;

    return;
    }

/************************************************************************/
/*									*/
/*  Cause subsequent drawing to be done in a certain color.		*/
/*									*/
/*  NOTE that background drawing with color= 0 is against the RTF idea	*/
/*	of the default background color: transparent.			*/
/*									*/
/************************************************************************/

void docDrawSetColorRgb(	DrawingContext *	dc,
				void *			through,
				int			r,
				int			g,
				int			b )
    {
    if  ( ! dc->dcCurrentColorSet		||
	  dc->dcCurrentColor.rgb8Red != r	||
	  dc->dcCurrentColor.rgb8Green != g	||
	  dc->dcCurrentColor.rgb8Blue != b	)
	{
	if  ( ! dc->dcSetColorRgb				||
	      (*dc->dcSetColorRgb)( dc, through, r, g, b )	)
	    { LLLDEB(r,g,b); return;	}

	dc->dcCurrentColor.rgb8Red= r;
	dc->dcCurrentColor.rgb8Green= g;
	dc->dcCurrentColor.rgb8Blue= b;
	dc->dcCurrentColorSet= 1;
	}

    return;
    }

void docDrawSetColorNumber(	DrawingContext *	dc,
				void *			through,
				int			colorNumber )
    {
    const BufferDocument *	bd= dc->dcDocument;
    const DocumentProperties *	dp= &(bd->bdProperties);
    RGB8Color			rgb8;

    rgb8.rgb8Red= 0;
    rgb8.rgb8Green= 0;
    rgb8.rgb8Blue= 0;

    if  ( colorNumber != 0 )
	{
	if  ( colorNumber < 1			||
	      colorNumber >= dp->dpColorCount	)
	    { LLDEB(colorNumber,dp->dpColorCount);	}
	else{ rgb8= dp->dpColors[colorNumber];		}
	}

    docDrawSetColorRgb( dc, through,
				rgb8.rgb8Red,
				rgb8.rgb8Green,
				rgb8.rgb8Blue );

    return;
    }


/************************************************************************/
/*									*/
/*  Generic Document drawing machinery.					*/
/*									*/
/************************************************************************/

static int docDrawOneLine(	void *				through,
				const BufferItem *		bi,
				int				part,
				const ParagraphFrame *		pf,
				int				line,
				DrawingContext *		dc )
    {
    const TextLine *		tl= bi->biParaLines+ line;

    int				accepted;

    const BorderProperties * const	bpLeft= (const BorderProperties *)0;
    const BorderProperties * const	bpRight= (const BorderProperties *)0;

    if  ( part == 0				&&
	  bi->biParaBorderAboveParagraph	&&
	  dc->dcDrawParaTop			)
	{
	LayoutPosition		lpTop;

	lpTop= bi->biTopPosition;
	lpTop.lpPageYTwips += bi->biParaSpaceBeforeTwips;

	if  ( (*dc->dcDrawParaTop)( bi->biParaBorderAboveParagraph,
						    bpLeft, bpRight,
						    pf, &lpTop, through, dc ) )
	    { LDEB(1); return -1;	}
	}

    if  ( dc->dcDrawTextLine )
	{
	accepted= (*dc->dcDrawTextLine)( bi, line, pf, through, dc );
	}
    else{ accepted= tl->tlParticuleCount;	}

    if  ( accepted < 1 )
	{ LDEB(accepted); return -1;	}

    if  ( part+ accepted == bi->biParaParticuleCount )
	{
	if  ( bi->biParaBorderBelowParagraph	&&
	      dc->dcDrawParaBottom		)
	    {
	    LayoutPosition	lpBelow;

	    lpBelow= bi->biBelowPosition;
	    lpBelow.lpPageYTwips -= bi->biParaSpaceAfterTwips;

	    if  ( (*dc->dcDrawParaBottom)( bi->biParaBorderBelowParagraph,
						bpLeft, bpRight,
						pf, &lpBelow, through, dc ) )
		{ LDEB(1); return -1;	}
	    }
	}

    return accepted;
    }

static int docDrawTextLines(	void *				through,
				const BufferItem *		bi,
				const LayoutPosition *		lpHere,
				const ParagraphFrame *		pf,
				int *				pLinesDone,
				DrawingContext *		dc )
    {
    int				done= 0;
    int				pastSelectionEnd= 0;

    int				line= *pLinesDone;
    const TextLine *		tl= bi->biParaLines+ line;

    while( line < bi->biParaLineCount )
	{
	int	accepted;
	int	beforeSelectionBegin= 0;

	if  ( tl->tlTopPosition.lpPage > lpHere->lpPage )
	    { break;	}

	if  ( tl->tlTopPosition.lpPage < lpHere->lpPage )
	    { beforeSelectionBegin= 1;	}

	if  ( dc->dcClipRect )
	    {
	    AppDrawingData *		add= dc->dcDrawingData;

	    int	y1= TL_BELOW_PIXELS( add, tl );

	    if  ( TL_TOP_PIXELS( add, tl ) > dc->dcClipRect->drY1 )
		{ pastSelectionEnd= 1;	}
	    if  ( y1 < dc->dcClipRect->drY0 )
		{ beforeSelectionBegin= 1;	}
	    }

	if  ( dc->dcFirstPage >= 0			&&
	      lpHere->lpPage < dc->dcFirstPage	)
	    { beforeSelectionBegin= 1;	}

	if  ( dc->dcDocumentSelection					&&
	      ! beforeSelectionBegin					&&
	      ! pastSelectionEnd					&&
	      ( bi == dc->dcDocumentSelection->dsBegin.dpBi	||
	        bi == dc->dcDocumentSelection->dsEnd.dpBi	)	)
	    {
	    const DocumentSelection *	ds= dc->dcDocumentSelection;

	    DocumentSelection		dsLine;
	    const int			mindPart= 1;

	    docLineSelection( &dsLine, bi, line );

	    if  ( bi == ds->dsEnd.dpBi					&&
		  docComparePositions( &(dsLine.dsBegin),
				    &(ds->dsEnd), mindPart ) > 0	)
		{ pastSelectionEnd= 1;	}

	    if  ( bi == ds->dsBegin.dpBi				&&
		  docComparePositions( &(dsLine.dsEnd),
				    &(ds->dsBegin), mindPart ) < 0	)
		{ beforeSelectionBegin= 1;	}
	    }

	if  ( beforeSelectionBegin || pastSelectionEnd )
	    { accepted= tl->tlParticuleCount;		}
	else{
	    accepted= docDrawOneLine( through, bi, tl->tlFirstParticule,
							    pf, line, dc );
	    }

	if  ( accepted < 1 )
	    { LDEB(accepted); return -1; }

	done += accepted; line++; tl++;
	}

    *pLinesDone= line;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Skip to the next page.						*/
/*									*/
/*  1)  Just to make the compiler happy, if nothing is to be printed	*/
/*	anymore, the current position is not going to be used.		*/
/*									*/
/************************************************************************/

static int docDrawToNextColumn(	BufferItem *		thisBodyBi,
				BufferItem *		nextBodyBi,
				void *			through,
				LayoutPosition *	lpHere,
				BlockFrame *		bf,
				DrawingContext *	dc )
    {
    BufferDocument *		bd= dc->dcDocument;

    const int			asLast= 0;

    int				startNewPage= 0;

    while( thisBodyBi && thisBodyBi->biLevel != DOClevSECT )
	{ thisBodyBi= thisBodyBi->biParent; }
    while( nextBodyBi && nextBodyBi->biLevel != DOClevSECT )
	{ nextBodyBi= nextBodyBi->biParent; }

    if  ( dc->dcFirstPage < 0				||
	  lpHere->lpPage >= dc->dcFirstPage		)
	{
	startNewPage= 1;

	if  ( thisBodyBi						&&
	      dc->dcDrawHeadersFooters					)
	    {
	    if  ( docDrawFootnotesForColumn( lpHere->lpPage, through, dc ) )
		{ LDEB(lpHere->lpPage); return -1;	}

	    if  ( docDrawPageFooter( thisBodyBi, through, dc, lpHere->lpPage ) )
		{ LDEB(lpHere->lpPage); return -1;	}
	    }

	if  ( dc->dcFinishPage						&&
	      (*dc->dcFinishPage)( through, dc, lpHere->lpPage,
							    asLast )	)
	    { LDEB(1); return -1;	}
	}

    if  ( dc->dcLastPage < 0				||
	  lpHere->lpPage+ 1 <= dc->dcLastPage		)
	{
	lpHere->lpPage++;

	if  ( nextBodyBi )
	    {
	    const SectionProperties *	sp= &(nextBodyBi->biSectProperties);
	    const DocumentGeometry *	dg= &(sp->spDocumentGeometry);

	    docLayoutSectColumnTop( nextBodyBi, bd, lpHere, bf );

	    if  ( startNewPage )
		{
		if  ( dc->dcStartPage					&&
		      (*dc->dcStartPage)( through, dg, dc,
						    lpHere->lpPage )	)
		    { LDEB(1); return -1;	}

		if  ( dc->dcDrawHeadersFooters				&&
		      docDrawPageHeader( nextBodyBi, through, dc,
						    lpHere->lpPage )	)
		    { LDEB(lpHere->lpPage); return -1;	}
		}
	    }
	}

    return 0;
    }

int docDrawToPageOfItem(		BufferItem *		prevBodyBi,
					BufferItem *		thisBodyBi,
					BufferItem *		thisBi,
					void *			through,
					LayoutPosition *	lpHere,
					DrawingContext *	dc )
    {
    BlockFrame	bf;

    while( thisBi->biTopPosition.lpPage > lpHere->lpPage	&&
	   thisBi->biTopPosition.lpPage > dc->dcFirstPage	)
	{
	if  ( docDrawToNextColumn( prevBodyBi, thisBodyBi, through,
						    lpHere, &bf, dc )	)
	    { SDEB(docLevelStr(thisBi->biLevel)); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Count the number of lines that fit on the current page.		*/
/*									*/
/************************************************************************/

static int docDrawBelowPara(	int *				pLinesDone,
				const BufferItem *		paraBi,
				int				countAfter,
				int				linesDone,
				const LayoutPosition *		lpHere,
				LayoutPosition *		lpBelow )
    {
    const TextLine *	tl;
    LayoutPosition	lp;


    tl= paraBi->biParaLines+ linesDone;
    while( linesDone < paraBi->biParaLineCount		&&
	   tl->tlTopPosition.lpPage <= lpHere->lpPage	)
	{
	if  ( tl->tlTopPosition.lpPage < lpHere->lpPage )
	    { linesDone++; tl++; continue;	}

	lp.lpPage= tl->tlTopPosition.lpPage;
	lp.lpPageYTwips= tl->tlTopPosition.lpPageYTwips+
					    tl->tlLineSpacingTwips;
	lp.lpAtTopOfColumn= 0;

	docLayoutPushBottomDown( lpBelow, &lp );

	linesDone++; tl++;
	}

    if  ( linesDone == paraBi->biParaLineCount )
	{
	lp= paraBi->biBelowPosition;

	if  ( ! countAfter )
	    { lp.lpPageYTwips -= paraBi->biParaSpaceAfterTwips;	}

	docLayoutPushBottomDown( lpBelow, &lp );
	}

    *pLinesDone= linesDone;
    return 0;
    }

/************************************************************************/
/*									*/
/*  Print a node in the BufferItem hierarchy.				*/
/*									*/
/*  1)  The last page is finished by the caller.			*/
/*									*/
/************************************************************************/

static int docDrawParaItem(	BufferItem *			paraBi,
				void *				through,
				DrawingContext *		dc )
    {
    ParagraphFrame		pf;
    int				line= 0;

    int				bottomTwips= -1;
    int				stripHigh= -1;

    BufferDocument *		bd= dc->dcDocument;

    BlockFrame			bf;
    LayoutPosition		lpTop;
    LayoutPosition		lpShadeTop;

    docBlockFrameTwips( &bf, paraBi, bd, paraBi->biTopPosition.lpPage,
					    paraBi->biTopPosition.lpColumn );

    docParagraphFrameTwips( &pf, &bf, bottomTwips, stripHigh, paraBi );

    if  ( dc->dcParaFramePixels )
	{ (*dc->dcParaFramePixels) ( &pf, dc->dcDrawingData, paraBi ); }

    lpTop= paraBi->biTopPosition;
    lpShadeTop= lpTop;

    if  ( ! lpShadeTop.lpAtTopOfColumn )
	{ lpShadeTop.lpPageYTwips += paraBi->biParaSpaceBeforeTwips; }

    while( line < paraBi->biParaLineCount )
	{
	if  ( dc->dcDrawParaShade )
	    {
	    int			paraLinesDone= line;
	    LayoutPosition	lpBelow;
	    const int		countAfter= 0;

	    lpBelow= lpTop;

	    if  ( docDrawBelowPara( &paraLinesDone, paraBi, countAfter, line,
							&lpTop, &lpBelow ) )
		{ LDEB(1); return -1;	}

	    if  ( paraLinesDone > line )
		{
		if  ( (*dc->dcDrawParaShade)( &(paraBi->biParaProperties),
						through, dc,
						pf.pfX0GeometryTwips,
						pf.pfX1GeometryTwips,
						&lpShadeTop, &lpBelow ) )
		    { LDEB(1); return -1;	}
		}
	    }

	if  ( docDrawTextLines( through, paraBi, &lpTop, &pf, &line, dc ) )
	    { LDEB(lpTop.lpPage); return -1;	}

	/*  1  */
	if  ( dc->dcLastPage >= 0		&&
	      lpTop.lpPage >= dc->dcLastPage	)
	    { break;	}

	if  ( line < paraBi->biParaLineCount )
	    {
	    if  ( docDrawToNextColumn( paraBi, paraBi,
					    through, &lpTop, &bf, dc )	)
		{ SDEB(docLevelStr(paraBi->biLevel));return -1;	}

	    lpShadeTop= lpTop;
	    }
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw part of the grid originating from a table row.			*/
/*									*/
/*  2)  Get the table row below the current row.			*/
/*  3)  Draw the cell shades (If any)					*/
/*  4)  Draw the cells.							*/
/*  5)  Do not draw the borders that cross cells with a colspan.	*/
/*  6)  If a cell has a right boder, and the next one a left border,	*/
/*	MS-Word only draws the right border of the leftmost cell.	*/
/*	( The same applies in the vertical direction. )			*/
/*  7)  Base the decisions on whether to draw a table grid on the same	*/
/*	criteria.							*/
/*  8)  Draw cell top border.						*/
/*  9)  Draw cell left border.						*/
/* 10)  Draw cell right border.						*/
/* 11)  Draw cell bottom border.					*/
/*									*/
/************************************************************************/

static int docDrawCellOrnaments(
				const BufferItem *		rowBi,
				const CellDrawingProgress *	cdpRow,
				int				atTopOfPage,
				int				atBottomOfPage,
				void *				through,
				DrawingContext *		dc,
				const LayoutPosition *		lpTop,
				const LayoutPosition *		lpBelow )
    {
    const CellDrawingProgress *	cdp;
    const CellProperties *	cp;

    int				col;

    /*  4  */
    cdp= cdpRow;
    cp= rowBi->biRowCells;
    for ( col= 0; col < rowBi->biChildCount; cdp++, cp++, col++ )
	{
	int				topAsGrid= 0;
	int				leftAsGrid= 0;
	int				rightAsGrid= 0;
	int				bottomAsGrid= 0;

	int				drawCellLeft= 1;
	int				drawCellRight= 1;

	const BorderProperties *	bpLeft;
	const BorderProperties *	bpRight;
	const BorderProperties *	bpTop;
	const BorderProperties *	bpBottom;

	int				useAbove= 0;
	int				useBelow= 0;
	int				leftHasBorder= 0;
	int				rightHasBorder= 0;

	docGetCellTopBorder( &bpTop, &useAbove,
					    rowBi, col, atTopOfPage );
	docGetCellBottomBorder( &bpBottom, &useBelow,
					    rowBi, col, atBottomOfPage );

	/*  5  */
	if  ( cp->cpMergedWithLeft )
	    { drawCellLeft= 0;	}
	if  ( col < rowBi->biChildCount- 1	&&
	      cp[1].cpMergedWithLeft		)
	    { drawCellRight= 0;	}

	/*  6  */
	if  ( col > 0 && cp[-1].cpRightBorder.bpStyle != DOCbsNONE )
	    { leftHasBorder= 1;	}
	if  ( col < rowBi->biChildCount- 1		&&
	      cp[1].cpLeftBorder.bpStyle != DOCbsNONE	)
	    { rightHasBorder= 1;	}

	bpLeft= &(cp->cpLeftBorder);
	bpRight= &(cp->cpRightBorder);

	if  ( leftHasBorder )
	    { bpLeft= &(cp[-1].cpRightBorder);		}
	if  ( cp->cpRightBorder.bpStyle == DOCbsNONE && rightHasBorder )
	    { bpRight= &(cp[1].cpLeftBorder);	}

	/*  7  */
	if  ( dc->dcDrawTableGrid )
	    {
	    if  ( cp->cpLeftBorder.bpStyle == DOCbsNONE		&&
		  ! leftHasBorder				)
		{ leftAsGrid= 1;	}

	    if  ( cp->cpRightBorder.bpStyle == DOCbsNONE	&&
		  ! rightHasBorder				)
		{ rightAsGrid= 1;	}

	    if  ( bpTop->bpStyle == DOCbsNONE		&&
		  ! useAbove				)
		{ topAsGrid= 1;	}

	    if  ( bpBottom->bpStyle == DOCbsNONE	&&
		  ! useBelow				)
		{ bottomAsGrid= 1;	}
	    }

	/*  3  */
	if  ( dc->dcDrawCellShade )
	    {
	    if  ( (*dc->dcDrawCellShade)( cp, bpTop, bpLeft, bpRight,
					through, dc,
					cdp->cdpX0Twips, cdp->cdpX1Twips,
					lpTop, lpBelow ) )
		{ LDEB(1); return -1;	}
	    }

	/*  8  */
	if  ( dc->dcDrawCellTop					&&
	      ! useAbove					&&
	      ( bpTop->bpStyle != DOCbsNONE || topAsGrid )	)
	    {
	    if  ( (*dc->dcDrawCellTop)(		bpTop, bpLeft, bpRight,
						topAsGrid,
						cdp->cdpX0Twips,
						cdp->cdpX1Twips,
						through, dc, lpTop ) )
		{ LDEB(1); return -1;	}
	    }

	/*  9  */
	if  ( drawCellLeft						&&
	      dc->dcDrawCellLeft					&&
	      ! leftHasBorder						&&
	      ( cp->cpLeftBorder.bpStyle != DOCbsNONE || leftAsGrid )	)
	    {
	    if  ( (*dc->dcDrawCellLeft)(	&(cp->cpLeftBorder),
						bpTop, bpBottom,
						leftAsGrid,
						through, dc,
						cdp->cdpX0Twips,
						lpTop, lpBelow ) )
		{ LDEB(1); return -1;	}
	    }

	/*  10  */
	if  ( drawCellRight						&&
	      dc->dcDrawCellRight					&&
	      ( cp->cpRightBorder.bpStyle != DOCbsNONE || rightAsGrid )	)
	    {
	    if  ( (*dc->dcDrawCellRight)(	&(cp->cpRightBorder),
						bpTop, bpBottom,
						rightAsGrid,
						through, dc,
						cdp->cdpX1Twips,
						lpTop, lpBelow ) )
		{ LDEB(1); return -1;	}
	    }

	/*  11  */
	if  ( dc->dcDrawCellBottom					&&
	      ( cp->cpBottomBorder.bpStyle != DOCbsNONE || bottomAsGrid ) )
	    {
	    if  ( (*dc->dcDrawCellBottom)(	bpBottom, bpLeft, bpRight,
						bottomAsGrid,
						cdp->cdpX0Twips,
						cdp->cdpX1Twips,
						through, dc, lpBelow ) )
						/*
						&(rowBi->biBelowPosition) ) )
						*/
		{ LDEB(1); return -1;	}
	    }
	}

    return 0;
    }


/************************************************************************/
/*									*/
/*  Draw the lines in a table cell that appear on one printed page.	*/
/*									*/
/************************************************************************/

static int docDrawBelowCell(	int *				pAdvanced,
				int *				pColToNextPage,
				const BufferItem *		cellBi,
				const LayoutPosition *		lpHere,
				const CellDrawingProgress *	cdp,
				LayoutPosition *		lpBelow )
    {
    int		para= cdp->cdpPara;
    int		linesDone= cdp->cdpLinesDone;

    int		advanced= 0;

    while( para < cellBi->biChildCount )
	{
	BufferItem *		paraBi= cellBi->biChildren[para];
	int			paraLinesDone= linesDone;
	const int		countAfter= 1;

	if  ( docDrawBelowPara( &paraLinesDone, paraBi, countAfter, linesDone,
							    lpHere, lpBelow ) )
	    { LDEB(1); return -1;	}

	if  ( paraLinesDone != linesDone )
	    { advanced= 1; linesDone= paraLinesDone;	}

	if  ( linesDone < paraBi->biParaLineCount )
	    { break;	}

	linesDone= 0; para++;
	}

    if  ( para < cellBi->biChildCount )
	{ *pColToNextPage= 1;	}
    if  ( advanced )
	{ *pAdvanced= 1;	}

    return 0;
    }

static int docDrawAdvanceCell(	void *				through,
				const BufferItem *		cellBi,
				const LayoutPosition *		lpHere,
				CellDrawingProgress *		cdp,
				BlockFrame *			bf,
				DrawingContext *		dc )
				
    {
    int				bottomTwips= -1;
    int				stripHigh= -1;

    int				advanced= 0;
    LayoutPosition		lpTop;
    LayoutPosition		lpShadeTop;

    lpTop= *lpHere;
    lpShadeTop= lpTop;

    while( cdp->cdpPara < cellBi->biChildCount )
	{
	BufferItem *		paraBi= cellBi->biChildren[cdp->cdpPara];
	int			linesDone= cdp->cdpLinesDone;

	if  ( paraBi->biTopPosition.lpPage > lpHere->lpPage )
	    { break;	}

	if  ( cdp->cdpLinesDone == 0 )
	    {
	    docParagraphFrameTwips( &(cdp->cdpParagraphFrame), bf,
					    bottomTwips, stripHigh, paraBi );

	    if  ( dc->dcParaFramePixels )
		{
		(*dc->dcParaFramePixels) ( &(cdp->cdpParagraphFrame),
						dc->dcDrawingData, paraBi );
		}

	    lpTop= paraBi->biTopPosition;
	    lpShadeTop= lpTop;
	    lpShadeTop.lpPageYTwips += paraBi->biParaSpaceBeforeTwips;
	    }

	if  ( dc->dcDrawParaShade )
	    {
	    int			paraLinesDone= cdp->cdpLinesDone;
	    LayoutPosition	lpBelow;
	    const int		countAfter= 0;

	    lpBelow= lpTop;

	    if  ( docDrawBelowPara( &paraLinesDone, paraBi, countAfter,
					cdp->cdpLinesDone, &lpTop, &lpBelow ) )
		{ LDEB(1); return -1;	}

	    if  ( paraLinesDone > cdp->cdpLinesDone )
		{
		if  ( (*dc->dcDrawParaShade)( &(paraBi->biParaProperties),
				    through, dc,
				    cdp->cdpParagraphFrame.pfX0GeometryTwips,
				    cdp->cdpParagraphFrame.pfX1GeometryTwips,
				    &lpShadeTop, &lpBelow ) )
		    { LDEB(1); return -1;	}
		}
	    }

	if  ( docDrawTextLines( through, paraBi, lpHere,
				&(cdp->cdpParagraphFrame), &linesDone, dc ) )
	    { LDEB(lpHere->lpPage); return -1;	}

	if  ( linesDone > cdp->cdpLinesDone )
	    { advanced= 1;	}

	cdp->cdpLinesDone= linesDone;

	if  ( cdp->cdpLinesDone < paraBi->biParaLineCount )
	    { break;	}

	cdp->cdpLinesDone= 0;
	cdp->cdpPara++;
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Print as much of a table row as fits on the current page.		*/
/*									*/
/************************************************************************/

static int docDrawRowPageStrip(	BufferItem *			rowBi,
				const LayoutPosition *		lpHere,
				int *				pToNextPage,
				CellDrawingProgress *		cdpRow,
				void *				through,
				DrawingContext *		dc,
				const LayoutPosition *		lpTop )
    {
    BufferDocument *		bd= dc->dcDocument;
    int				col;

    int				toNextPage= 0;

    int				rowAdvanced;
    int				atRowTop= 1;

    LayoutPosition		lpBottom;

    CellDrawingProgress *	cdp;

    BlockFrame			bf;

    docBlockFrameTwips( &bf, rowBi, bd, lpHere->lpPage, lpHere->lpColumn );

    lpBottom= *lpTop;

    rowAdvanced= 0;
    cdp= cdpRow;
    for ( col= 0; col < rowBi->biChildCount; cdp++, col++ )
	{
	BufferItem *	cellBi= rowBi->biChildren[col];
	int		colToNextPage= 0;

	cdp->cdpAdvanceColumnThisPage= 0;

	if  ( cdp->cdpColumnAdvanced )
	    { atRowTop= 0;	}

	if  ( docDrawBelowCell( &(cdp->cdpAdvanceColumnThisPage),
						&colToNextPage, cellBi,
						lpHere, cdp, &lpBottom ) )
	    { LDEB(1); return -1;	}

	if  ( cdp->cdpAdvanceColumnThisPage )
	    { rowAdvanced= 1; cdp->cdpColumnAdvanced= 1;	}

	if  ( colToNextPage )
	    { toNextPage= 1;	}
	}

    if  ( toNextPage == 0 )
	{
	docLayoutPushBottomDown( &lpBottom, &(rowBi->biBelowPosition) );
	}

    if  ( rowAdvanced )
	{
	const int	atTopOfColumn= lpTop->lpAtTopOfColumn;
	const int	atBottomOfPage= toNextPage;

	if  ( docDrawCellOrnaments( rowBi, cdpRow,
					atTopOfColumn, atBottomOfPage,
					through, dc, lpTop, &lpBottom ) )
	    { LDEB(1); return -1;	}
	}

    rowAdvanced= 0;
    cdp= cdpRow;
    for ( col= 0; col < rowBi->biChildCount; cdp++, col++ )
	{
	BufferItem *	cellBi= rowBi->biChildren[col];
	int		cellAdvanced= 0;
	int		colToNextPage= 0;

	if  ( dc->dcDocumentSelection )
	    {
	    const DocumentSelection *	ds= dc->dcDocumentSelection;

	    if  ( docCompareItemPositions( cellBi, ds->dsBegin.dpBi ) < 0 )
		{ continue;	}

	    if  ( docCompareItemPositions( cellBi, ds->dsEnd.dpBi ) > 0	)
		{ continue;	}

	    if  ( ( ds->dsCol0 >= 0 && col < ds->dsCol0	)	||
	          ( ds->dsCol1 >= 0 && col > ds->dsCol1	)	)
		{ continue;	}
	    }

	if  ( cdp->cdpColumnAdvanced )
	    { atRowTop= 0;	}

	if  ( docDrawBelowCell( &cellAdvanced, &colToNextPage, cellBi,
						lpHere, cdp, &lpBottom ) )
	    { LDEB(1); return -1;	}

	if  ( docDrawAdvanceCell( through, cellBi, lpHere, cdp, &bf, dc ) )
	    { LDEB(1); return -1;	}

	if  ( cellAdvanced )
	    { rowAdvanced= 1; cdp->cdpColumnAdvanced= 1;	}

	if  ( cdp->cdpPara < cellBi->biChildCount )
	    { toNextPage= 1;	}
	}

    *pToNextPage= toNextPage; return 0;
    }

/************************************************************************/
/*									*/
/*  Draw a table row by successively drawing the portions that fit on	*/
/*  the current page and moving to the next page.			*/
/*									*/
/*  1)  The last page is finished by the caller.			*/
/*									*/
/************************************************************************/

static int docDrawRowItem(	BufferItem *			rowBi,
				void *				through,
				DrawingContext *		dc )
    {
    AppDrawingData *		add= dc->dcDrawingData;
    double			xfac= add->addMagnifiedPixelsPerTwip;

    int				col;

    const RowProperties *	rp= &(rowBi->biRowProperties);
    BufferItem *		sectBi= rowBi->biParent;
    const DocumentGeometry *	dg= &(sectBi->biSectDocumentGeometry);
    CellProperties *		cp;

    int				toNextPage;

    static CellDrawingProgress *	colP;
    CellDrawingProgress *		cdp;

    int				x00Pixels;

    int				x0Twips;
    int				x1Twips;
    int				x1Pixels;

    LayoutPosition		lpTop;
    LayoutPosition		lpHere;

    if  ( rp->rpCellCount < rowBi->biChildCount )
	{
	LLDEB(rp->rpCellCount,rowBi->biChildCount);
	docListItem( 0, rowBi );
	return -1;
	}

    cdp= (CellDrawingProgress *)realloc( colP,
				rp->rpCellCount* sizeof(CellDrawingProgress) );
    if  ( ! cdp )
	{ LXDEB(rp->rpCellCount,cdp); return -1;	}
    colP= cdp;

    toNextPage= 0;

    x00Pixels= TWIPStoPIXELS( xfac, dg->dgLeftMarginTwips );

    x0Twips= dg->dgLeftMarginTwips+ rowBi->biRowLeftIndentTwips;

    cp= rp->rpCells;
    for ( col= 0; col < rowBi->biChildCount; cp++, col++ )
	{
	x1Twips= dg->dgLeftMarginTwips+ cp->cpRightBoundaryTwips;
	x1Pixels= TWIPStoPIXELS( xfac, x1Twips );

	cdp[col].cdpColumnAdvanced= 0;
	cdp[col].cdpAdvanceColumnThisPage= 0;

	cdp[col].cdpPara= 0;
	cdp[col].cdpLinesDone= 0;

	cdp[col].cdpX0Twips= x0Twips;
	cdp[col].cdpX1Twips= x1Twips;

	x0Twips= x1Twips;
	}

    lpTop= rowBi->biTopPosition;
    lpHere= rowBi->biTopPosition;

    if  ( docDrawRowPageStrip( rowBi, &lpHere, &toNextPage, cdp,
						    through, dc, &lpTop ) )
	{ LDEB(1); return -1;	}

    while( toNextPage )
	{
	BlockFrame		bf;

	/*  1  */
	if  ( dc->dcLastPage >= 0		&&
	      lpHere.lpPage >= dc->dcLastPage	)
	    { break;	}

	if  ( docDrawToNextColumn( rowBi, rowBi, through, &lpTop, &bf, dc ) )
	    { SDEB(docLevelStr(rowBi->biLevel)); return -1;	}

	lpHere= lpTop;

	if  ( docDrawRowPageStrip( rowBi, &lpHere, &toNextPage, cdp,
						    through, dc, &lpTop ) )
	    { LDEB(1); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw the children of a group item.					*/
/*									*/
/************************************************************************/

static int docDrawGroupItem(	const BufferItem *		bi,
				void *				through,
				DrawingContext *		dc )
    {
    AppDrawingData *		add= dc->dcDrawingData;

    int				i;
    LayoutPosition		lpHere;

    BufferItem *		prevBi= (BufferItem *)0;

    if  ( bi->biChildCount > 0 )
	{ prevBi= bi->biChildren[0];	}

    lpHere= bi->biTopPosition;
    for ( i= 0; i < bi->biChildCount; i++ )
	{
	BufferItem *	thisBi= bi->biChildren[i];

	if  ( dc->dcClipRect						&&
	      BI_TOP_PIXELS( add, thisBi ) > dc->dcClipRect->drY1	)
	    { break;	}

	if  ( dc->dcDocumentSelection					&&
	      docCompareItemPositions( thisBi,
			    dc->dcDocumentSelection->dsEnd.dpBi ) > 0	)
	    { break;	}

	if  ( dc->dcLastPage >= 0				&&
	      thisBi->biTopPosition.lpPage > dc->dcLastPage	)
	    { break;	}

	if  ( dc->dcFirstPage < 0				||
	      thisBi->biBelowPosition.lpPage >= dc->dcFirstPage	)
	    {
	    if  ( docDrawToPageOfItem( prevBi, thisBi, thisBi,
						    through, &lpHere, dc ) )
		{ SDEB(docLevelStr(thisBi->biLevel)); return -1;	}

	    if  ( docDrawItem( thisBi, through, dc ) )
		{ LDEB(i); return -1;	}

	    lpHere= thisBi->biBelowPosition;
	    }

	prevBi= thisBi;
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw the document root item.					*/
/*									*/
/************************************************************************/

static int docDrawDocItem(	const BufferItem *		docBi,
				void *				through,
				DrawingContext *		dc )
    {
    BufferDocument *		bd= dc->dcDocument;
    DocumentProperties *	dp= &(bd->bdProperties);

    if  ( docDrawGroupItem( docBi, through, dc ) )
	{ LDEB(1); return -1;	}

    if  ( docBi->biInExternalItem == DOCinBODY				&&
	  dp->dpEndnoteProperties.npPosition == DPftnPOS_DOC_END	)
	{
	if  ( docDrawEndnotesForDocument( through, dc ) )
	    { LDEB(1); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw a section and its dependants.					*/
/*									*/
/************************************************************************/

static int docDrawSectItem(	const BufferItem *		sectBi,
				void *				through,
				DrawingContext *		dc )
    {
    BufferDocument *		bd= dc->dcDocument;
    DocumentProperties *	dp= &(bd->bdProperties);

    if  ( docDrawGroupItem( sectBi, through, dc ) )
	{ LDEB(1); return -1;	}

    if  ( sectBi->biInExternalItem == DOCinBODY				&&
	  dp->dpEndnoteProperties.npPosition == DPftnPOS_SECT_END	)
	{
	if  ( docDrawEndnotesForSection( sectBi->biNumberInParent,
							    through, dc ) )
	    { LDEB(1); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw an arbitrary item.						*/
/*									*/
/************************************************************************/

int docDrawItem(	BufferItem *			bi,
			void *				through,
			DrawingContext *		dc )
    {
    AppDrawingData *		add= dc->dcDrawingData;
    const DocumentSelection *	ds= dc->dcDocumentSelection;

    if  ( dc->dcClipRect )
	{
	if  ( BI_TOP_PIXELS( add, bi ) > dc->dcClipRect->drY1 )
	    { return 0;	}

	if  ( BI_BELOW_PIXELS( add, bi ) < dc->dcClipRect->drY0 )
	    { return 0; }
	}

    if  ( ds )
	{
	if  ( docCompareItemPositions( bi, ds->dsBegin.dpBi ) < 0 )
	    { return 0;	}

	if  ( docCompareItemPositions( bi, ds->dsEnd.dpBi ) > 0 )
	    { return 0; }
	}

    if  ( dc->dcFirstPage >= 0					&&
	  bi->biBelowPosition.lpPage < dc->dcFirstPage		)
	{ return 0;	}

    if  ( dc->dcLastPage >= 0					&&
	  bi->biTopPosition.lpPage > dc->dcLastPage		)
	{ return 0;	}

    switch( bi->biLevel )
	{
	case DOClevDOC:
	    if  ( docDrawDocItem( bi, through, dc ) )
		{ LDEB(1); return -1;	}
	    break;

	case DOClevSECT:
	    if  ( docDrawSectItem( bi, through, dc ) )
		{ LDEB(1); return -1;	}
	    break;

	case DOClevCELL:
	rowAsGroup:
	    if  ( docDrawGroupItem( bi, through, dc ) )
		{ LDEB(1); return -1;	}
	    break;

	case DOClevROW:
	    if  ( ! bi->biRowHasTableParagraphs )
		{ goto rowAsGroup;	}

	    if  ( docDrawRowItem( bi, through, dc ) )
		{ LDEB(1); return -1;	}

	    break;

	case DOClevPARA:
	    if  ( docDrawParaItem( bi, through, dc ) )
		{ LDEB(1); return -1;	}

	    break;

	default:
	    LDEB(bi->biLevel); return -1;
	}

    return 0;
    }
