/************************************************************************/
/*									*/
/*  Table editing.							*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<ctype.h>

#   include	"docEdit.h"
#   include	"docLayout.h"

#   include	<appDebugon.h>

/************************************************************************/
/*									*/
/*  Insert a series of rows in a table.					*/
/*									*/
/************************************************************************/

int docInsertTableRows(		BufferItem *		sectBi,
				BufferItem **		pNewBi,
				const BufferItem *	refRowBi,
				const RowProperties *	rp,
				int			textAttributeNumber,
				int			n,
				int			paraNr,
				int			rows,
				EditOperation *		eo )
    {
    DocumentPosition	dp;

    int			row;
    int			firstPos= n;

    for ( row= 0; row < rows; n++, row++ )
	{
	BufferItem *		rowBi= (BufferItem *)0;

	const int		stroffFrom= 0;
	const int		sectShift= 0;
	const int		stroffShift= 0;
	const int		paraShift= rp->rpCellCount;

	rowBi= docInsertRowItem( eo->eoBd, sectBi, n, rp, textAttributeNumber );
	if  ( ! rowBi )
	    { XDEB(rowBi); return -1;	}

	if  ( sectBi->biInExternalItem == DOCinBODY )
	    {
	    docShiftNoteReferences( eo->eoBd, paraNr, stroffFrom,
				    sectShift, paraShift, stroffShift );
	    }

	docEditShiftReformatRangeParaNr( eo, paraNr+ 1, paraShift );

	paraNr += rp->rpCellCount;

	if  ( refRowBi						&&
	      docCopyRowColumnAttributes( rowBi, refRowBi )	)
	    { LDEB(1); return -1;	}

	docEditIncludeItemInReformatRange( eo, rowBi );
	}

    docSectDelimitTables( sectBi );

    if  ( docFirstPosition( &dp, sectBi->biChildren[firstPos] ) )
	{ LDEB(firstPos);	}

    *pNewBi= dp.dpBi; return 0;
    }

/************************************************************************/
/*									*/
/*  Insert cells/columns in tables.					*/
/*									*/
/************************************************************************/

static BufferItem * docInsertCellInRow(
				BufferDocument *	bd,
				BufferItem *		rowBi,
				int			col,
				const CellProperties *	cp,
				int			textAttributeNumber )
    {
    const int * const	colorMap= (const int *)0;
    RowProperties *	rp;

    BufferItem *	cellBi;
    BufferItem *	paraBi;

    rp= &(rowBi->biRowProperties);

    if  ( docInsertRowColumn( rp, col, cp, colorMap ) )
	{ LDEB(col); return (BufferItem *)0;	}

    cellBi= docInsertItem( bd, rowBi, col, DOClevCELL );
    if  ( ! cellBi )
	{ LXDEB(col,cellBi); return (BufferItem *)0;	}

    paraBi= docInsertEmptyParagraph( bd, cellBi, textAttributeNumber );
    if  ( ! paraBi )
	{ XDEB(paraBi); return (BufferItem *)0;	}

    paraBi->biParaInTable= 1;

    return paraBi;
    }

/************************************************************************/
/*									*/
/*  Insert cells/columns in tables.					*/
/*									*/
/************************************************************************/

int docSplitColumnInRows(	BufferItem **		pNewParaBi,
				EditOperation *		eo,
				BufferItem *		sectBi,
				int			row0,
				int			row,
				int			row1,
				int			col,
				int			after )
    {
    BufferDocument *		bd= eo->eoBd;

    BufferItem *		newParaBi= (BufferItem *)0;

    BufferItem *		rowBi;

    if  ( after )
	{ after= 1;	}

    while( row0 <= row1 )
	{
	RowProperties *		rp;

	CellProperties *	oldCp;
	CellProperties		newCp;

	int			left;
	int			middle;

	BufferItem *		cellBi;
	BufferItem *		paraBi;

	int			textAttributeNumber;

	rowBi= sectBi->biChildren[row0];
	rp= &(rowBi->biRowProperties);

	oldCp= rp->rpCells+ col;

	if  ( col == 0 )
	    { left= rp->rpLeftIndentTwips;			}
	else{ left= rp->rpCells[col-1].cpRightBoundaryTwips;	}

	middle= ( rp->rpCells[col].cpRightBoundaryTwips+ left )/ 2;

	newCp= *oldCp;
	if  ( after )
	    { oldCp->cpRightBoundaryTwips= middle;	}
	else{ newCp. cpRightBoundaryTwips= middle;	}

	cellBi= rowBi->biChildren[col];
	paraBi= cellBi->biChildren[0];
	textAttributeNumber= paraBi->biParaParticules[0].tpTextAttributeNumber;

	paraBi= docInsertCellInRow( bd, rowBi, col+ after, &newCp,
							textAttributeNumber );
	if  ( row0 == row )
	    { newParaBi= paraBi;	}

	if  ( sectBi->biInExternalItem == DOCinBODY )
	    {
	    const int	stroffFrom= 0;
	    const int	stroffShift= 0;
	    const int	sectShift= 0;

	    const int	paraNr= docNumberOfParagraph( paraBi );
	    const int	paraShift= 1;

	    docShiftNoteReferences( eo->eoBd, paraNr, stroffFrom,
				    sectShift, paraShift, stroffShift );
	    }

	docEditIncludeItemInReformatRange( eo, rowBi );

	row0++;
	}

    docSectDelimitTables( sectBi );

    *pNewParaBi= newParaBi;
    return 0;
    }

/************************************************************************/
/*									*/
/*  Update the properties of a table in response to a user action.	*/
/*									*/
/************************************************************************/

void docChangeTableLayout(	EditOperation *		eo,
				BufferItem *		sectBi,
				int			row0,
				int			row1,
				int			col0,
				int			col1,
				const PropertyMask *	rpSetMask,
				const PropertyMask *	cpSetMask,
				const RowProperties *	rpSet )
    {
    BufferItem *		rowBi= sectBi->biChildren[row0];

    PropertyMask		rpDoneMask;
    PropertyMask		cpDoneMask;

    const int * const		colorMap= (const int *)0;

    PROPmaskCLEAR( &rpDoneMask );
    PROPmaskCLEAR( &cpDoneMask );

    if  ( col1 > rpSet->rpCellCount- 1 )
	{ col1=  rpSet->rpCellCount- 1;	}

    while( row0 <= row1 )
	{
	int			col;

	rowBi= sectBi->biChildren[row0];

	if  ( docUpdRowProperties( &rpDoneMask, &(rowBi->biRowProperties),
						rpSetMask, rpSet, colorMap ) )
	    { LDEB(row0);	}

	if  ( ! PROPmaskISEMPTY( cpSetMask ) )
	    {
	    const CellProperties *	cpSet;
	    CellProperties *		cpTo;

	    cpSet= rpSet->rpCells+ col0;
	    cpTo= rowBi->biRowProperties.rpCells+ col0;
	    for ( col= col0; col <= col1; cpTo++, cpSet++, col++ )
		{
		if  ( col >= rowBi->biRowCellCount )
		    { break;	}

		if  ( docUpdCellProperties( &cpDoneMask, cpTo,
						cpSetMask, cpSet, colorMap ) )
		    { LLDEB(row0,col);	}
		}
	    }

	docEditIncludeItemInReformatRange( eo, rowBi );

	row0++;
	}

    docSectDelimitTables( sectBi );

    return;
    }

/************************************************************************/
/*									*/
/*  Delete columns from a table.					*/
/*									*/
/************************************************************************/

int docDeleteColumnsFromRows(	EditOperation *	eo,
				BufferItem *	sectBi,
				int		delRow0,
				int		delRow1,
				int		delCol0,
				int		delCol1 )
    {
    int				count= delCol1- delCol0+ 1;
    int				row;

    for ( row= delRow0; row <= delRow1; row++ )
	{
	BufferItem *	rowBi= sectBi->biChildren[row];

	int		firstParaDeleted= -1;
	int		paragraphsDeleted= 0;

	if  ( docDeleteColumnsFromRow( &(rowBi->biRowProperties),
							    delCol0, count ) )
	    { LLDEB(delCol0,count); continue;	}

	docEditDeleteItems( eo, &firstParaDeleted, &paragraphsDeleted,
						    rowBi, delCol0, count );

	if  ( rowBi->biInExternalItem == DOCinBODY )
	    {
	    const int		sectShift= 0;
	    const int		stroffFrom= 0;
	    const int		stroffShift= 0;

	    docShiftNoteReferences( eo->eoBd,
			    firstParaDeleted+ paragraphsDeleted, stroffFrom,
			    sectShift, -paragraphsDeleted, stroffShift );
	    }


	docEditIncludeItemInReformatRange( eo, rowBi );
	}

    docSectDelimitTables( sectBi );

    return 0;
    }

