/************************************************************************/
/*									*/
/*  Layout of a table row in the document.				*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stddef.h>
#   include	<stdio.h>
#   include	<stdlib.h>

#   define	y0	math_y0
#   define	y1	math_y1
#   include	<math.h>
#   undef	y0
#   undef	y1

#   include	"docLayout.h"

#   include	<appDebugon.h>

/************************************************************************/
/*									*/
/*  Calculate layout insets for a (table) row.				*/
/*									*/
/*  This routine is also used to calculate the inset that is used to	*/
/*  accomodate space for the bottom border of the previous row. This	*/
/*  inset is not only used in real rows, but also in the immediate	*/
/*  successor of a real row.						*/
/*									*/
/*  1)  Reserve space for the top border of all cells.			*/
/*									*/
/************************************************************************/

static void docLayoutCalculateRowTopInset(
					BufferItem *		rowBi,
					int			atTopOfPage )
    {
    int				col;

    rowBi->biTopInsetTwips= 0;

    if  ( ! rowBi->biRowHasTableParagraphs )
	{ LDEB(rowBi->biRowHasTableParagraphs); return;	}

    /*  1  */
    for ( col= 0; col < rowBi->biChildCount; col++ )
	{
	int				useAbove= 0;
	const BorderProperties *	bpTop;

	docGetCellTopBorder( &bpTop, &useAbove, rowBi, col, atTopOfPage );

	docLayoutStretchInsetForBorder( &(rowBi->biTopInsetTwips), bpTop );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Layout as much of a table row as fits on the current page.		*/
/*									*/
/************************************************************************/

static int docLayoutRowPageStrip( BufferItem *			rowBi,
				int *				pToNextPage,
				int *				pRowHeight,
				const ParagraphLayoutContext *	plc,
				ParagraphLayoutJob *		plj,
				BlockFrame *			bf,
				LayoutJob *			lj )
    {
    int				col;

    int				continueOnNextPage= 0;

    LayoutPosition		lpBottom;
    int				bottomTwips= -1;
    int				stripHigh= -1;
    const CellProperties *	cp;

    const ScreenLayout *	sl= &(plc->plcScreenLayout);

    lpBottom= lj->ljPosition;

    continueOnNextPage= 0;

    if  ( rowBi->biRowHeightTwips < 0 )
	{
	bottomTwips= lj->ljPosition.lpPageYTwips-
				    rowBi->biRowHeightTwips- *pRowHeight;

	stripHigh= -rowBi->biRowHeightTwips;
	}

    cp= rowBi->biRowCells;
    for ( col= 0; col < rowBi->biChildCount; cp++, plj++, col++ )
	{
	BufferItem *			cellBi= rowBi->biChildren[col];
	ParagraphLayoutPosition *	plp= &(plj->pljPos);

	if  ( cp->cpMergedWithLeft )
	    { continue;	}

	if  ( plp->plpProgress.pspPara >= cellBi->biChildCount )
	    { continue;	}

	plp->plpPos= lj->ljPosition;

	docParagraphStripFrame( sl, cellBi, plc->plcAdd,
					bottomTwips, stripHigh, bf, plj );

	if  ( docLayoutParagraphsInStrip( plj, bf, plc, cellBi,
			    plj->pljParaUpto, bottomTwips, stripHigh ) )
	    { LDEB(1); return -1;	}

	docLayoutPushBottomDown( &lpBottom, &(plp->plpPos) );

	if  ( cellBi->biInExternalItem == DOCinBODY		&&
	      plp->plpProgress.pspPara < plj->pljParaUpto  	&&
	      ( bottomTwips < 0				||
	        plp->plpPos.lpPageYTwips < bottomTwips	)	)
	    { continueOnNextPage= 1;	}
	}

    *pRowHeight += lpBottom.lpPageYTwips- lj->ljPosition.lpPageYTwips;

    lj->ljPosition= lpBottom;

    *pToNextPage= continueOnNextPage;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Calculate the layout of a row in the document.			*/
/*									*/
/************************************************************************/

static void docLayoutStartRow(	BufferItem *			rowBi,
				LayoutPosition *		lp,
				ParagraphLayoutJob *		plj )
    {
    CellProperties *		cp;
    int				col;

    if  ( rowBi->biTopInsetTwips > 0 )
	{
	lp->lpPageYTwips += rowBi->biTopInsetTwips;
	lp->lpAtTopOfColumn= 0;
	}

    cp= rowBi->biRowCells;
    for ( col= 0; col < rowBi->biChildCount; cp++, plj++, col++ )
	{
	BufferItem *	cellBi= rowBi->biChildren[col];

	docPsInitParagraphLayoutJob( plj );

	plj->pljParaUpto= cellBi->biChildCount;

	docPsBeginParagraphLayoutProgress( plj, 0, 0, 0,
						cellBi->biChildCount, lp );

	cellBi->biTopPosition= *lp;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Continue a table row on the next page.				*/
/*									*/
/*  1)  Finish the footnotes that we found on the current page.		*/
/*  2)  Move to the next page. (Column in multi column sections)	*/
/*  3)  Reserve space for the top border of the cells.			*/
/*  4)  Initialize the layout of the different table columns.		*/
/*									*/
/************************************************************************/

static int docRowToNextPage(	BufferItem *			rowBi,
				LayoutJob *			lj,
				int				bottomTwips,
				int				stripHigh,
				BlockFrame *			bf,
				ParagraphLayoutJob *		rowPlj,
				const ParagraphLayoutContext *	plc )
    {
    CellProperties *		cp;
    int				col;

    ParagraphLayoutJob *	plj;

    LayoutPosition		lpBefore;

    const int			atTopOfPage= 1;
    int				inset= 0;

    lpBefore= lj->ljPosition;

    /*  1  */
    if  ( BF_HAS_FOOTNOTES( bf )					&&
	  docLayoutFootnotesForColumn( bf, &(lj->ljPosition), lj )	)
	{ LDEB(1); return -1;	}

    /*  2  */
    docLayoutToNextColumn( rowBi, lj->ljBd, &(lj->ljPosition), bf );

    /*  3  */
    for ( col= 0; col < rowBi->biChildCount; col++ )
	{
	int				useAbove= 0;
	const BorderProperties *	bpTop;

	docGetCellTopBorder( &bpTop, &useAbove, rowBi, col, atTopOfPage );

	docLayoutStretchInsetForBorder( &inset, bpTop );
	}

    lj->ljPosition.lpPageYTwips += inset;

    /*  4  */
    cp= rowBi->biRowCells;
    plj= rowPlj;
    for ( col= 0; col < rowBi->biChildCount; cp++, plj++, col++ )
	{
	BufferItem *		cellBi= rowBi->biChildren[col];
	ParagraphStripPosition	psp0;
	int			advanced;

	if  ( plj->pljPos.plpProgress.pspPara >= cellBi->biChildCount )
	    { continue;	}

	psp0= plj->pljPos0.plpProgress;

	docPsAdvanceParagraphLayout( &advanced,
					    &psp0,
					    &(plj->pljPos0.plpProgress),
					    &(plj->pljPos.plpProgress),
					    lpBefore.lpPage, cellBi );

	if  ( advanced )
	    {
	    plj->pljPos0.plpProgress= psp0;
	    plj->pljPos.plpProgress= psp0;
	    }
	else{ plj->pljPos0.plpProgress= plj->pljPos.plpProgress;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Finish row layout.							*/
/*									*/
/*  1)  The bottom is below all cells.					*/
/*  2)  If there is a bottom inset, step over it.			*/
/*  3)  Stretch height to the minimum height (If any)			*/
/*  3)  Stretch height to the exact given height (If any)		*/
/*									*/
/************************************************************************/

static void docLayoutFinishRow(	BufferItem *		rowBi,
				LayoutJob *		lj,
				ParagraphLayoutJob *	plj,
				int			rowHeightTwips )
    {
    int				col;
    CellProperties *		cp;

    /*  1  */
    cp= rowBi->biRowCells;
    for ( col= 0; col < rowBi->biChildCount; cp++, plj++, col++ )
	{
	BufferItem *			cellBi= rowBi->biChildren[col];
	ParagraphLayoutPosition *	plp= &(plj->pljPos);

	if  ( cp->cpMergedWithLeft )
	    { continue;	}

	plp->plpPos.lpPageYTwips += cellBi->biBottomInsetTwips;
	plp->plpPos.lpAtTopOfColumn= 0;

	cellBi->biBelowPosition= plp->plpPos;

	docLayoutPushBottomDown( &(lj->ljPosition), &(plp->plpPos) );
	}

    /*  2  */
    if  ( rowBi->biBottomInsetTwips > 0 )
	{
	lj->ljPosition.lpPageYTwips += rowBi->biBottomInsetTwips;
	lj->ljPosition.lpAtTopOfColumn= 0;
	}

    /*  3  */
    if  ( rowHeightTwips < rowBi->biRowHeightTwips )
	{
	lj->ljPosition.lpPageYTwips +=
				rowBi->biRowHeightTwips- rowHeightTwips;
	lj->ljPosition.lpAtTopOfColumn= 0;
	}

    /*  4  */
    if  ( rowHeightTwips < -rowBi->biRowHeightTwips )
	{
	lj->ljPosition.lpPageYTwips +=
				-rowBi->biRowHeightTwips- rowHeightTwips;
	lj->ljPosition.lpAtTopOfColumn= 0;
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Calculate the layout of a table row.				*/
/*									*/
/*  1)  Sanity check against crashes.					*/
/*  2)  Allocate memory to monitor the progress of the layout of the	*/
/*	cells in the row.						*/
/*  3)  Initialize and place as much row content on the current page	*/
/*	as possible.							*/
/*  4)  Some of the content of the row did not fit on the current page.	*/
/*	Find out whether the row must be restarted on the next page.	*/
/*	(This does not make sense if the top of the row already is at	*/
/*	the top of the page.)						*/
/*  5)  When there are cells that remain empty on this page.. restart	*/
/*	on the next page.						*/
/*  6)  Go to the next page.						*/
/*  7)  If there is a reason, reinitialize the row layout job, to move	*/
/*	the whole row to the next page.					*/
/*  8)  Continue or restart (7) on the next page.			*/
/*  9)  As long as necessary, move to the next page and place content	*/
/*	there. (The exceptions above only apply for the first page, as	*/
/*	it is clearly impossible to keep the contents of a table higher	*/
/*	than one page on one page)					*/
/*									*/
/************************************************************************/

int docLayoutRowItem(		BufferItem *			rowBi,
				BlockFrame *			bf,
				LayoutJob *			lj,
				const ParagraphLayoutContext *	plc )
    {
    int				rowHeightTwips;
    int				toNextPage;

    static ParagraphLayoutJob *	staticPlj;
    ParagraphLayoutJob *	fresh;

    int				bottomTwips= -1;

    const ScreenLayout *	sl= &(plc->plcScreenLayout);

    /*  1  */
    if  ( rowBi->biRowCellCount < rowBi->biChildCount )
	{
	LLDEB(rowBi->biRowCellCount,rowBi->biChildCount);
	docListItem( 0, rowBi );
	rowBi->biChildCount= rowBi->biRowCellCount;
	}

    /*  2  */
    fresh= (ParagraphLayoutJob *)realloc( staticPlj,
		    rowBi->biRowCellCount* sizeof(ParagraphLayoutJob) );
    if  ( ! fresh )
	{ LXDEB(rowBi->biRowCellCount,fresh); return -1;	}
    staticPlj= fresh;

    if  ( sl && sl->slStartRow					&&
	  (*sl->slStartRow)( rowBi, lj->ljAdd, lj->ljBd )	)
	{ LDEB(1); return -1;	}

    /*  3  */
    toNextPage= 0;
    rowHeightTwips= 0;
    if  ( rowBi->biRowHeightTwips < 0 )
	{ bottomTwips= lj->ljPosition.lpPageYTwips- rowBi->biRowHeightTwips; }

    docLayoutCalculateRowTopInset( rowBi, lj->ljPosition.lpAtTopOfColumn );

    docLayoutStartRow( rowBi, &(lj->ljPosition), staticPlj );

    if  ( docLayoutRowPageStrip( rowBi, &toNextPage,
				    &rowHeightTwips, plc, staticPlj, bf, lj ) )
	{ LDEB(1); return -1;	}

    /*  4  */
    if  ( toNextPage && ! rowBi->biTopPosition.lpAtTopOfColumn )
	{
	int				someAtTop= 0;
	int				col;
	const ParagraphLayoutJob *	plj;

	/*  5  */
	plj= staticPlj;
	for ( col= 0; col < rowBi->biChildCount; plj++, col++ )
	    {
	    const ParagraphLayoutPosition *	plp= &(plj->pljPos);
	    const ParagraphStripPosition *	psp= &(plp->plpProgress);

	    if  ( psp->pspPara == 0 && psp->pspPart == 0 )
		{
		const BufferItem *	cellBi= rowBi->biChildren[col];

		if  ( cellBi->biChildCount > 0 )
		    {
		    const BufferItem *	paraBi;

		    paraBi= cellBi->biChildren[psp->pspPara];

		    if  ( psp->pspPart < paraBi->biParaParticuleCount )
			{ someAtTop= 1; break; }
		    }
		}
	    }

	/*  6  */
	if  ( docRowToNextPage( rowBi, lj, bottomTwips, rowHeightTwips,
							bf, staticPlj, plc ) )
	    { LDEB(toNextPage); return -1;	}

	/*  7  */
	if  ( rowBi->biRowKeepOnOnePage || someAtTop )
	    {
	    rowBi->biTopPosition= lj->ljPosition;

	    docLayoutCalculateRowTopInset( rowBi,
					    lj->ljPosition.lpAtTopOfColumn );

	    docLayoutStartRow( rowBi, &(lj->ljPosition), staticPlj );

	    rowHeightTwips= 0;
	    if  ( rowBi->biRowHeightTwips < 0 )
		{
		bottomTwips=
			lj->ljPosition.lpPageYTwips- rowBi->biRowHeightTwips;
		}
	    }

	/*  8  */
	if  ( docLayoutRowPageStrip( rowBi, &toNextPage,
				&rowHeightTwips, plc, staticPlj, bf, lj ) )
	    { LDEB(1); return -1;	}
	}

    /*  9  */
    while( toNextPage )
	{
	if  ( docRowToNextPage( rowBi, lj, bottomTwips, rowHeightTwips,
							bf, staticPlj, plc ) )
	    { LDEB(toNextPage); return -1;	}

	if  ( docLayoutRowPageStrip( rowBi, &toNextPage,
				&rowHeightTwips, plc, staticPlj, bf, lj ) )
	    { LDEB(1); return -1;	}
	}

    docLayoutFinishRow( rowBi, lj, staticPlj, rowHeightTwips );

    return 0;
    }

