/************************************************************************/
/*									*/
/*  Manage paragraph properties.					*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>

#   include	<appDebugon.h>

#   include	<appUnit.h>

#   include	"docParaProperties.h"

/************************************************************************/
/*									*/
/*  Initialise a tab stop.						*/
/*									*/
/************************************************************************/

void docInitTabStop(	TabStop *	ts )
    {
    ts->tsTwips= 0;
    ts->tsPixels= 0;
    ts->tsAlignment= DOCtaLEFT;
    ts->tsLeader= DOCtlNONE;
    }

/************************************************************************/
/*									*/
/*  Set a tab stop in a paragraph.					*/
/*									*/
/************************************************************************/

int docParaAddTab(		ParagraphProperties *	pp,
				const TabStop *		tsNew )
    {
    TabStop *		ts;
    int			newSize;
    int			i;

    if  ( pp->ppTabCount % 10 )
	{ newSize= pp->ppTabCount+  1; }
    else{ newSize= pp->ppTabCount+ 11; }

    newSize *= sizeof(TabStop);

    ts= (TabStop *)realloc( pp->ppTabStops, newSize );
    if  ( ! ts )
	{ LXDEB(newSize,ts); return -1; }
    pp->ppTabStops= ts;

    i= pp->ppTabCount; ts= pp->ppTabStops+ pp->ppTabCount;
    while( i > 0 && ts[-1].tsTwips > tsNew->tsTwips )
	{ ts[0]= ts[-1]; i--; ts--;	}

    *ts= *tsNew;
    pp->ppTabCount++;

    return i;
    }

/************************************************************************/
/*									*/
/*  Remove a tab stop from the ruler of a paragraph.			*/
/*									*/
/************************************************************************/

void docParaDeleteTab(	ParagraphProperties *		pp,
			int				n )
    {
    TabStop *		ts;

    if  ( n < 0 || n >= pp->ppTabCount )
	{ LLDEB(n,pp->ppTabCount); return;	}

    pp->ppTabCount--;

    ts= pp->ppTabStops+ n;
    while( n < pp->ppTabCount )
	{ ts[0]= ts[1]; ts++; n++;	}

    return;
    }

/************************************************************************/
/*									*/
/*  Manage paragraph properties.					*/
/*									*/
/************************************************************************/

void docCleanParagraphProperties(	ParagraphProperties *	pp )
    {
    if  ( pp->ppTabStops )
	{ free( pp->ppTabStops );	}
    }

void docInitParagraphProperties(	ParagraphProperties *	pp )
    {
    pp->ppTabCount= 0;
    pp->ppTabStops= (TabStop *)0;

    pp->ppFirstIndentTwips= 0;
    pp->ppLeftIndentTwips= 0;
    pp->ppRightIndentTwips= 0;

    pp->ppSpaceBeforeTwips= 0;
    pp->ppSpaceAfterTwips= 0;
    pp->ppLineSpacingTwips= 0;

    pp->ppStyle= 0;

    docInitBorderProperties( &(pp->ppTopBorder) );
    docInitBorderProperties( &(pp->ppLeftBorder) );
    docInitBorderProperties( &(pp->ppRightBorder) );
    docInitBorderProperties( &(pp->ppBottomBorder) );
    docInitBorderProperties( &(pp->ppBoxBorder) );
    docInitBorderProperties( &(pp->ppBetweenBorder) );
    docInitBorderProperties( &(pp->ppBar) );

    docInitItemShading( &(pp->ppShading) );

    pp->ppOutlineLevel= 9;

    pp->ppAlignment= DOCiaLEFT;

    pp->ppStartsOnNewPage= 0;
    pp->ppInTable= 0;
    pp->ppLineSpacingIsMultiple= 0;

    pp->ppKeepOnPage= 0;
    pp->ppKeepWithNext= 0;
    pp->ppWidowControl= 0;
    pp->ppHyphenateParagraph= 0;

    return;
    }

/************************************************************************/
/*									*/
/*  Change paragraph properties and tell what has been changed.		*/
/*									*/
/************************************************************************/

int docUpdParaProperties(	PropertyMask *			pPpChgPask,
				ParagraphProperties *		ppTo,
				const PropertyMask *		ppUpdMask,
				const ParagraphProperties *	ppFrom,
				const int *			colorMap )
    {
    PropertyMask		ppChgMask;
    PropertyMask		isUpdMask;

    PROPmaskCLEAR( &ppChgMask );

    if  ( PROPmaskISSET( ppUpdMask, PPpropSTYLE ) )
	{
	if  ( ppTo->ppStyle != ppFrom->ppStyle )
	    {
	    ppTo->ppStyle= ppFrom->ppStyle;
	    PROPmaskADD( &ppChgMask, PPpropSTYLE );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropIN_TABLE ) )
	{
	if  ( ppTo->ppInTable != ppFrom->ppInTable )
	    {
	    ppTo->ppInTable= ppFrom->ppInTable;
	    PROPmaskADD( &ppChgMask, PPpropIN_TABLE );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropFIRST_INDENT ) )
	{
	if  ( ppTo->ppFirstIndentTwips != ppFrom->ppFirstIndentTwips )
	    {
	    ppTo->ppFirstIndentTwips= ppFrom->ppFirstIndentTwips;
	    PROPmaskADD( &ppChgMask, PPpropFIRST_INDENT );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLEFT_INDENT ) )
	{
	if  ( ppTo->ppLeftIndentTwips != ppFrom->ppLeftIndentTwips )
	    {
	    ppTo->ppLeftIndentTwips= ppFrom->ppLeftIndentTwips;
	    PROPmaskADD( &ppChgMask, PPpropLEFT_INDENT );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropRIGHT_INDENT ) )
	{
	if  ( ppTo->ppRightIndentTwips != ppFrom->ppRightIndentTwips )
	    {
	    ppTo->ppRightIndentTwips= ppFrom->ppRightIndentTwips;
	    PROPmaskADD( &ppChgMask, PPpropRIGHT_INDENT );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropALIGNMENT ) )
	{
	if  ( ppTo->ppAlignment != ppFrom->ppAlignment )
	    {
	    ppTo->ppAlignment= ppFrom->ppAlignment;
	    PROPmaskADD( &ppChgMask, PPpropALIGNMENT );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropTAB_STOPS ) )
	{
	if  ( ppTo->ppTabCount > 0 || ppFrom->ppTabCount > 0 )
	    {
	    if  ( ppTo->ppTabCount != ppFrom->ppTabCount )
		{ PROPmaskADD( &ppChgMask, PPpropTAB_STOPS );	}

	    if  ( ppFrom->ppTabCount == 0 )
		{ ppTo->ppTabCount= 0;	}
	    else{
		TabStop *			ts;
		TabStop *			ots;
		TabStop *			nts;
		int				tab;

		ts= (TabStop *)malloc( ppFrom->ppTabCount* sizeof(TabStop) );
		if  ( ! ts )
		    { LXDEB(ppFrom->ppTabCount,ts); return -1;	}

		ots= ppTo->ppTabStops;
		nts= ppFrom->ppTabStops;
		for ( tab= 0; tab < ppFrom->ppTabCount; ots++, nts++, tab++ )
		    {
		    if  ( ppTo->ppTabCount == ppFrom->ppTabCount )
			{
			if  ( nts->tsTwips != ots->tsTwips		||
			      nts->tsAlignment != ots->tsAlignment	||
			      nts->tsLeader != ots->tsLeader		)
			    {
			    PROPmaskADD( &ppChgMask, PPpropTAB_STOPS );
			    ts[tab]= *nts;
			    }
			else{ ts[tab]= *ots;	}
			}
		    else{ ts[tab]= *nts;		}
		    }

		ppTo->ppTabCount= ppFrom->ppTabCount;
		if  ( ppTo->ppTabStops )
		    { free( ppTo->ppTabStops );	}

		ppTo->ppTabStops= ts;
		}
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropNEWPAGE ) )
	{
	if  ( ppTo->ppStartsOnNewPage != ppFrom->ppStartsOnNewPage )
	    {
	    ppTo->ppStartsOnNewPage= ppFrom->ppStartsOnNewPage;
	    PROPmaskADD( &ppChgMask, PPpropNEWPAGE );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropWIDCTLPAR ) )
	{
	if  ( ppTo->ppWidowControl != ppFrom->ppWidowControl )
	    {
	    ppTo->ppWidowControl= ppFrom->ppWidowControl;
	    PROPmaskADD( &ppChgMask, PPpropWIDCTLPAR );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropKEEPN ) )
	{
	if  ( ppTo->ppKeepWithNext != ppFrom->ppKeepWithNext )
	    {
	    ppTo->ppKeepWithNext= ppFrom->ppKeepWithNext;
	    PROPmaskADD( &ppChgMask, PPpropKEEPN );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropKEEP ) )
	{
	if  ( ppTo->ppKeepOnPage != ppFrom->ppKeepOnPage )
	    {
	    ppTo->ppKeepOnPage= ppFrom->ppKeepOnPage;
	    PROPmaskADD( &ppChgMask, PPpropKEEP );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropSPACE_BEFORE ) )
	{
	if  ( ppTo->ppSpaceBeforeTwips != ppFrom->ppSpaceBeforeTwips )
	    {
	    ppTo->ppSpaceBeforeTwips= ppFrom->ppSpaceBeforeTwips;
	    PROPmaskADD( &ppChgMask, PPpropSPACE_BEFORE );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropSPACE_AFTER ) )
	{
	if  ( ppTo->ppSpaceAfterTwips != ppFrom->ppSpaceAfterTwips )
	    {
	    ppTo->ppSpaceAfterTwips= ppFrom->ppSpaceAfterTwips;
	    PROPmaskADD( &ppChgMask, PPpropSPACE_AFTER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLINE_SPACING_DIST ) )
	{
	if  ( ppTo->ppLineSpacingTwips != ppFrom->ppLineSpacingTwips )
	    {
	    ppTo->ppLineSpacingTwips= ppFrom->ppLineSpacingTwips;
	    PROPmaskADD( &ppChgMask, PPpropLINE_SPACING_DIST );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLINE_SPACING_MULT ) )
	{
	if  ( ppTo->ppLineSpacingIsMultiple != ppFrom->ppLineSpacingIsMultiple )
	    {
	    ppTo->ppLineSpacingIsMultiple= ppFrom->ppLineSpacingIsMultiple;
	    PROPmaskADD( &ppChgMask, PPpropLINE_SPACING_MULT );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropTOP_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppTopBorder),
					&(ppFrom->ppTopBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppTopBorder),
					&(ppFrom->ppTopBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropTOP_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBOTTOM_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBottomBorder),
					&(ppFrom->ppBottomBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppBottomBorder),
					&(ppFrom->ppBottomBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropBOTTOM_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLEFT_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppLeftBorder),
					&(ppFrom->ppLeftBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppLeftBorder),
					&(ppFrom->ppLeftBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropLEFT_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropRIGHT_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppRightBorder),
					&(ppFrom->ppRightBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppRightBorder),
					&(ppFrom->ppRightBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropRIGHT_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBOX_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBoxBorder),
					&(ppFrom->ppBoxBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppBoxBorder),
					&(ppFrom->ppBoxBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropBOX_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBETWEEN_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBetweenBorder),
					&(ppFrom->ppBetweenBorder), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppBetweenBorder),
					&(ppFrom->ppBetweenBorder), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropBETWEEN_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBAR_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBar),
					&(ppFrom->ppBar), colorMap ) )
	    {
	    docCopyBorderProperties( &(ppTo->ppBar),
					&(ppFrom->ppBar), colorMap );
	    PROPmaskADD( &ppChgMask, PPpropBAR_BORDER );
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropOUTLINELEVEL ) )
	{
	if  ( ppTo->ppOutlineLevel != ppFrom->ppOutlineLevel )
	    {
	    ppTo->ppOutlineLevel= ppFrom->ppOutlineLevel;
	    PROPmaskADD( &ppChgMask, PPpropOUTLINELEVEL );
	    }
	}

    docShadingMaskFromParagraphMask( &isUpdMask, ppUpdMask );
    if  ( ! utilPropMaskIsEmpty( &isUpdMask ) )
	{
	PropertyMask	isChgMask;

	PROPmaskCLEAR( &isChgMask );

	docUpdateItemShading( &isChgMask, &(ppTo->ppShading),
				&isUpdMask, &(ppFrom->ppShading), colorMap );

	docShadingMaskToParagraphMask( &isChgMask, &isChgMask );
	utilPropMaskOr( &ppChgMask, &ppChgMask, &isChgMask );
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropHYPHPAR ) )
	{
	if  ( ppTo->ppHyphenateParagraph != ppFrom->ppHyphenateParagraph )
	    {
	    ppTo->ppHyphenateParagraph= ppFrom->ppHyphenateParagraph;
	    PROPmaskADD( &ppChgMask, PPpropHYPHPAR );
	    }
	}

    *pPpChgPask= ppChgMask; return 0;
    }

/************************************************************************/
/*									*/
/*  Compare paragraph properties.					*/
/*									*/
/************************************************************************/

void docParaPropertyDifference(	PropertyMask *			pChgMask,
				const ParagraphProperties *	ppTo,
				const ParagraphProperties *	ppFrom,
				const PropertyMask *		ppUpdMask )
    {
    PropertyMask		ppChgMask;
    PropertyMask		isUpdMask;
    const int *			colorMap= (const int *)0;

    PROPmaskCLEAR( &ppChgMask );

    if  ( PROPmaskISSET( ppUpdMask, PPpropSTYLE ) )
	{
	if  ( ppTo->ppStyle != ppFrom->ppStyle )
	    { PROPmaskADD( &ppChgMask, PPpropSTYLE ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropFIRST_INDENT ) )
	{
	if  ( ppTo->ppFirstIndentTwips != ppFrom->ppFirstIndentTwips )
	    { PROPmaskADD( &ppChgMask, PPpropFIRST_INDENT ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLEFT_INDENT ) )
	{
	if  ( ppTo->ppLeftIndentTwips != ppFrom->ppLeftIndentTwips )
	    { PROPmaskADD( &ppChgMask, PPpropLEFT_INDENT ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropRIGHT_INDENT ) )
	{
	if  ( ppTo->ppRightIndentTwips != ppFrom->ppRightIndentTwips )
	    { PROPmaskADD( &ppChgMask, PPpropRIGHT_INDENT ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropALIGNMENT ) )
	{
	if  ( ppTo->ppAlignment != ppFrom->ppAlignment )
	    { PROPmaskADD( &ppChgMask, PPpropALIGNMENT ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropTAB_STOPS ) )
	{
	if  ( ppTo->ppTabCount > 0 || ppFrom->ppTabCount > 0 )
	    {
	    if  ( ppTo->ppTabCount != ppFrom->ppTabCount )
		{ PROPmaskADD( &ppChgMask, PPpropTAB_STOPS );	}
	    else{
		TabStop *			ts1;
		TabStop *			ts2;
		int				tab;

		ts1= ppTo->ppTabStops;
		ts2= ppFrom->ppTabStops;
		for ( tab= 0; tab < ppFrom->ppTabCount; ts1++, ts2++, tab++ )
		    {
		    if  ( ts1->tsTwips != ts2->tsTwips			||
			  ts1->tsAlignment != ts2->tsAlignment		||
			  ts1->tsLeader != ts2->tsLeader		)
			{ PROPmaskADD( &ppChgMask, PPpropTAB_STOPS ); }
		    }
		}
	    }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropNEWPAGE ) )
	{
	if  ( ppTo->ppStartsOnNewPage != ppFrom->ppStartsOnNewPage )
	    { PROPmaskADD( &ppChgMask, PPpropNEWPAGE ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropWIDCTLPAR ) )
	{
	if  ( ppTo->ppWidowControl != ppFrom->ppWidowControl )
	    { PROPmaskADD( &ppChgMask, PPpropWIDCTLPAR ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropKEEPN ) )
	{
	if  ( ppTo->ppKeepWithNext != ppFrom->ppKeepWithNext )
	    { PROPmaskADD( &ppChgMask, PPpropKEEPN ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropKEEP ) )
	{
	if  ( ppTo->ppKeepOnPage != ppFrom->ppKeepOnPage )
	    { PROPmaskADD( &ppChgMask, PPpropKEEP ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropSPACE_BEFORE ) )
	{
	if  ( ppTo->ppSpaceBeforeTwips != ppFrom->ppSpaceBeforeTwips )
	    { PROPmaskADD( &ppChgMask, PPpropSPACE_BEFORE ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropSPACE_AFTER ) )
	{
	if  ( ppTo->ppSpaceAfterTwips != ppFrom->ppSpaceAfterTwips )
	    { PROPmaskADD( &ppChgMask, PPpropSPACE_AFTER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLINE_SPACING_DIST ) )
	{
	if  ( ppTo->ppLineSpacingTwips != ppFrom->ppLineSpacingTwips )
	    { PROPmaskADD( &ppChgMask, PPpropLINE_SPACING_DIST ); }
	}
    if  ( PROPmaskISSET( ppUpdMask, PPpropLINE_SPACING_MULT ) )
	{
	if  ( ppTo->ppLineSpacingIsMultiple != ppFrom->ppLineSpacingIsMultiple )
	    { PROPmaskADD( &ppChgMask, PPpropLINE_SPACING_MULT ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropTOP_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppTopBorder),
					&(ppFrom->ppTopBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropTOP_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBOTTOM_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBottomBorder),
					&(ppFrom->ppBottomBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropBOTTOM_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropLEFT_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppLeftBorder),
					&(ppFrom->ppLeftBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropLEFT_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropRIGHT_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppRightBorder),
					&(ppFrom->ppRightBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropRIGHT_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBOX_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBoxBorder),
					&(ppFrom->ppBoxBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropBOX_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBETWEEN_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBetweenBorder),
					&(ppFrom->ppBetweenBorder), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropBETWEEN_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropBAR_BORDER ) )
	{
	if  ( docBordersDiffer( &(ppTo->ppBar), &(ppFrom->ppBar), colorMap ) )
	    { PROPmaskADD( &ppChgMask, PPpropBAR_BORDER ); }
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropOUTLINELEVEL ) )
	{
	if  ( ppTo->ppOutlineLevel != ppFrom->ppOutlineLevel )
	    { PROPmaskADD( &ppChgMask, PPpropOUTLINELEVEL );	}
	}

    docShadingMaskFromParagraphMask( &isUpdMask, ppUpdMask );
    if  ( ! utilPropMaskIsEmpty( &isUpdMask ) )
	{
	PropertyMask	isChgMask;

	PROPmaskCLEAR( &isChgMask );

	docItemShadingDifference( &isChgMask, &(ppTo->ppShading),
				&(ppFrom->ppShading), &isUpdMask, colorMap );

	docShadingMaskToParagraphMask( &isChgMask, &isChgMask );
	utilPropMaskOr( &ppChgMask, &ppChgMask, &isChgMask );
	}

    if  ( PROPmaskISSET( ppUpdMask, PPpropHYPHPAR ) )
	{
	if  ( ppTo->ppHyphenateParagraph != ppFrom->ppHyphenateParagraph )
	    { PROPmaskADD( &ppChgMask, PPpropHYPHPAR ); }
	}

    *pChgMask= ppChgMask;
    return;
    }

/************************************************************************/
/*									*/
/*  Translate an item shading property mask to a paragraph property	*/
/*  mask								*/
/*									*/
/************************************************************************/

void docShadingMaskToParagraphMask(	PropertyMask *		ppPropMask,
					const PropertyMask *	isPropMask )
    {
    PropertyMask	pm;

    PROPmaskCLEAR( &pm );

    if  ( PROPmaskISSET( isPropMask, ISpropFORE_COLOR ) )
	{ PROPmaskADD( &pm, PPpropSHADE_FORE_COLOR );	}

    if  ( PROPmaskISSET( isPropMask, ISpropBACK_COLOR ) )
	{ PROPmaskADD( &pm, PPpropSHADE_BACK_COLOR );	}

    if  ( PROPmaskISSET( isPropMask, ISpropLEVEL ) )
	{ PROPmaskADD( &pm, PPpropSHADE_LEVEL );	}

    if  ( PROPmaskISSET( isPropMask, ISpropPATTERN ) )
	{ PROPmaskADD( &pm, PPpropSHADE_PATTERN );	}

    *ppPropMask= pm;
    return;
    }

void docShadingMaskFromParagraphMask(	PropertyMask *		isPropMask,
					const PropertyMask *	ppPropMask )
    {
    PropertyMask	pm;

    PROPmaskCLEAR( &pm );

    if  ( PROPmaskISSET( ppPropMask, PPpropSHADE_FORE_COLOR ) )
	{ PROPmaskADD( &pm, ISpropFORE_COLOR );	}

    if  ( PROPmaskISSET( ppPropMask, PPpropSHADE_BACK_COLOR ) )
	{ PROPmaskADD( &pm, ISpropBACK_COLOR );	}

    if  ( PROPmaskISSET( ppPropMask, PPpropSHADE_LEVEL ) )
	{ PROPmaskADD( &pm, ISpropLEVEL );	}

    if  ( PROPmaskISSET( ppPropMask, PPpropSHADE_PATTERN ) )
	{ PROPmaskADD( &pm, ISpropPATTERN );	}

    *isPropMask= pm;
    return;
    }

