/************************************************************************/
/*									*/
/*  Read the various document tables of an RTF text file into a		*/
/*  BufferDocument.							*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>
#   include	<ctype.h>

#   include	<appDebugon.h>

#   include	<appUnit.h>
#   include	<psFont.h>
#   include	"docRtf.h"

/************************************************************************/
/*									*/
/*  Consume a font table.						*/
/*									*/
/************************************************************************/

static int docRtfCommitFaltName(	RtfReadingContext *	rrc )
    {
    if  ( docRtfRememberText( &(rrc->rrcCurrentFont.dfAltName), rrc ) )
	{ LDEB(rrc->rrcSavedText.mbSize); return -1; }

    return 0;
    }

static int docRtfFontFaltGroup( SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc )
    {
    if  ( docRtfReadGroup( sis, rcw->rcwLevel,
				    (RtfControlWord *)0, 0, 0, rrc,
				    docRtfEmptyTable, docRtfEmptyTable,
				    docRtfSaveText, docRtfCommitFaltName ) )
	{ SLDEB(rcw->rcwWord,arg); return -1;	}

    return 0;
    }

static int docRtfFontPanoseText(	RtfReadingContext *	rrc,
					const unsigned char *	panose,
					int			len )
    {
    if  ( len != FONTlenPANOSE )
	{ LLDEB(len,FONTlenPANOSE); return 0;	}

    memcpy( rrc->rrcCurrentFont.dfPanose, panose, len );

    rrc->rrcCurrentFont.dfPanose[FONTlenPANOSE]= '\0';

    return 0;
    }

static int docRtfFontPanoseGroup( SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc )
    {
    if  ( docRtfReadGroup( sis, rcw->rcwLevel,
				    (RtfControlWord *)0, 0, 0, rrc,
				    docRtfEmptyTable, docRtfEmptyTable,
				    docRtfFontPanoseText, (RtfCommitGroup)0 ) )
	{ SLDEB(rcw->rcwWord,arg); return -1;	}

    return 0;
    }

static int docRtfAddCurrentFontToList(	RtfReadingContext *	rrc )
    {
    if  ( docRtfRememberText( &(rrc->rrcCurrentFont.dfName), rrc ) )
	{ LDEB(rrc->rrcSavedText.mbSize); return -1; }

    if  ( rrc->rrcCurrentFont.dfName )
	{
	DocumentFont *		df;
	DocumentProperties *	dp= &(rrc->rrcBd->bdProperties);
	int			len;

	len= strlen( rrc->rrcCurrentFont.dfName );
	if  ( len > 0 )
	    {
	    if  ( rrc->rrcCurrentFont.dfName[len-1] == ';' )
		{ rrc->rrcCurrentFont.dfName[len-1]= '\0';	}
	    }

	df= docInsertFont( &(dp->dpFontList),
			rrc->rrcCurrentFont.dfDocFamilyNumber,
			rrc->rrcCurrentFont.dfFamilyStyle,
			rrc->rrcCurrentFont.dfName,
			rrc->rrcCurrentFont.dfAltName );

	if  ( ! df )
	    { SDEB(rrc->rrcCurrentFont.dfName); return -1; }

	df->dfCharset= rrc->rrcCurrentFont.dfCharset;
	df->dfPitch= rrc->rrcCurrentFont.dfPitch;

	strcpy( df->dfPanose, rrc->rrcCurrentFont.dfPanose );
	}

    docCleanFont( &(rrc->rrcCurrentFont) );
    docInitFont( &(rrc->rrcCurrentFont) );

    return 0;
    }

static int docRtfFontName(	RtfReadingContext *	rrc,
				const unsigned char *	name,
				int			len )
    {
    if  ( docRtfSaveText( rrc, name, len ) )
	{ LDEB(len); return -1;	}

    if  ( docRtfAddCurrentFontToList( rrc ) )
	{ LDEB(len); return -1;	}

    return 0;
    }

static int docRtfFontNumber(	SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc )
    {
    docCleanFont( &(rrc->rrcCurrentFont) );
    docInitFont( &(rrc->rrcCurrentFont) );

    rrc->rrcCurrentFont.dfDocFamilyNumber= arg;

    return 0;
    }

static int docRtfFontFamily(	SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc )
    {
    rrc->rrcCurrentFont.dfFamilyStyle= strdup( rcw->rcwWord );

    if  ( ! rrc->rrcCurrentFont.dfFamilyStyle )
	{ XDEB(rrc->rrcCurrentFont.dfFamilyStyle); return -1;	}

    return 0;
    }

static int docRtfFontProperty(	SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc )
    {
    switch( rcw->rcwId )
	{
	case RTFidFCHARSET:
	    rrc->rrcCurrentFont.dfCharset= arg;
	    break;

	case RTFidFPRQ:
	    rrc->rrcCurrentFont.dfPitch= arg;
	    break;

	default:
	    SLDEB(rcw->rcwWord,arg);
	}

    return 0;
    }

static RtfControlWord	docRtfFontGroupGroups[]=
    {
	{ "panose",	RTFidPANOSE,	DOClevANY, docRtfFontPanoseGroup, },
	{ "falt",	RTFidFALT,	DOClevANY, docRtfFontFaltGroup,	  },

	{ 0, 0, 0 }
    };

static RtfControlWord	docRtfFontGroupWords[]=
    {
	{ "fnil",	RTFidFNIL,	DOClevANY, docRtfFontFamily, },
	{ "froman",	RTFidFROMAN,	DOClevANY, docRtfFontFamily, },
	{ "fswiss",	RTFidFSWISS,	DOClevANY, docRtfFontFamily, },
	{ "fmodern",	RTFidFMODERN,	DOClevANY, docRtfFontFamily, },
	{ "fscript",	RTFidFSCRIPT,	DOClevANY, docRtfFontFamily, },
	{ "fdecor",	RTFidFDECOR,	DOClevANY, docRtfFontFamily, },
	{ "ftech",	RTFidFTECH,	DOClevANY, docRtfFontFamily, },

	{ "fcharset",	RTFidFCHARSET,	DOClevDOC, docRtfFontProperty, },
	{ "fprq",	RTFidFPRQ,	DOClevDOC, docRtfFontProperty, },

	{ 0, 0, 0 }
    };

static int docRtfFontGroup(	SimpleInputStream *	sis,
				const RtfControlWord *	rcw,
				int			arg,
				RtfReadingContext *	rrc	)
    {
    docCleanFont( &(rrc->rrcCurrentFont) );
    docInitFont( &(rrc->rrcCurrentFont) );

    rrc->rrcCurrentFont.dfDocFamilyNumber= arg;

    if  ( docRtfReadGroup( sis, rcw->rcwLevel,
				(RtfControlWord *)0, 0, 0, rrc,
				docRtfFontGroupWords, docRtfFontGroupGroups,
				docRtfSaveText, docRtfAddCurrentFontToList ) )
	{ SLDEB(rcw->rcwWord,arg); return -1;	}

    return 0;
    }

static RtfControlWord	docRtfFontTableWords[]=
    {
	{ "f",		RTFidF,		DOClevDOC, docRtfFontNumber, },

	{ "fnil",	RTFidFNIL,	DOClevANY, docRtfFontFamily, },
	{ "froman",	RTFidFROMAN,	DOClevANY, docRtfFontFamily, },
	{ "fswiss",	RTFidFSWISS,	DOClevANY, docRtfFontFamily, },
	{ "fmodern",	RTFidFMODERN,	DOClevANY, docRtfFontFamily, },
	{ "fscript",	RTFidFSCRIPT,	DOClevANY, docRtfFontFamily, },
	{ "fdecor",	RTFidFDECOR,	DOClevANY, docRtfFontFamily, },
	{ "ftech",	RTFidFTECH,	DOClevANY, docRtfFontFamily, },

	{ "fcharset",	RTFidFCHARSET,	DOClevDOC, docRtfFontProperty, },
	{ "fprq",	RTFidFPRQ,	DOClevDOC, docRtfFontProperty, },

	{ 0, 0, 0 }
    };

static RtfControlWord	docRtfFontTableGroups[]=
    {
	{ "f",		RTFidF,		DOClevDOC,  docRtfFontGroup, },

	{ 0, 0, 0 }
    };

int docRtfFontTable(	SimpleInputStream *	sis,
			const RtfControlWord *	rcw,
			int			arg,
			RtfReadingContext *	rrc	)
    {
    if  ( docRtfReadGroup( sis, rcw->rcwLevel,
				(RtfControlWord *)0, 0, 0, rrc,
				docRtfFontTableWords, docRtfFontTableGroups,
				docRtfFontName, (RtfCommitGroup)0 ) )
	{ SLDEB(rcw->rcwWord,arg); return -1;	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Write a font table.							*/
/*									*/
/************************************************************************/

void docRtfWriteFontTable(	SimpleOutputStream *		sos,
				int *				pCol,
				const DocumentFontList *	dfl )
    {
    int				i;
    const DocumentFont *	df= dfl->dflFonts;

    docRtfWriteDestinationBegin( "\\fonttbl", pCol, sos );
    docRtfWriteNextLine( pCol, sos );

    for ( i= 0; i < dfl->dflCount; df++, i++ )
	{
	int		afterTag= 0;

	if  ( ! df->dfFamilyStyle && ! df->dfName )
	    { continue;	}

	docRtfWriteArgDestinationBegin( "\\f", pCol,
					    df->dfDocFamilyNumber, sos );

	sioOutPutCharacter( '\\', sos ); *pCol += 1;
	if  ( df->dfFamilyStyle )
	    {
	    sioOutPutString( df->dfFamilyStyle, sos );
	    *pCol += strlen( df->dfFamilyStyle );
	    }
	else{
	    sioOutPutString( "fnil", sos );
	    *pCol += 4;
	    }
	afterTag= 1;

	if  ( df->dfCharset != FONTcharsetDEFAULT )
	    {
	    docRtfWriteArgTag( "\\fcharset", pCol, df->dfCharset, sos );
	    afterTag= 1;
	    }

	if  ( df->dfPitch != FONTpitchDEFAULT )
	    {
	    docRtfWriteArgTag( "\\fprq", pCol, df->dfPitch, sos );
	    afterTag= 1;
	    }

	if  ( df->dfPanose[0] )
	    {
	    const unsigned char * const outputMapping= (const unsigned char *)0;

	    docRtfWriteDestinationBegin( "\\*\\panose ", pCol, sos );

	    docRtfEscapeString( (unsigned char *)df->dfPanose,
					    outputMapping, pCol,
					    strlen( df->dfPanose ), sos );

	    docRtfWriteDestinationEnd( pCol, sos );
	    afterTag= 0;
	    }

	if  ( df->dfName )
	    {
	    if  ( afterTag )
		{ sioOutPutCharacter( ' ', sos ); *pCol += 1;	}

	    sioOutPutString( df->dfName, sos );
	    *pCol += strlen( df->dfName );
	    }

	if  ( df->dfAltName )
	    {
	    const unsigned char * const outputMapping= (const unsigned char *)0;

	    docRtfWriteDestinationBegin( "\\*\\falt ", pCol, sos );

	    docRtfEscapeString( (unsigned char *)df->dfAltName,
					    outputMapping, pCol,
					    strlen( df->dfAltName ), sos );

	    docRtfWriteDestinationEnd( pCol, sos );
	    afterTag= 0;
	    }

	sioOutPutCharacter( ';', sos ); *pCol += 1;

	docRtfWriteDestinationEnd( pCol, sos );
	if  ( i+ 1 < dfl->dflCount )
	    { docRtfWriteNextLine( pCol, sos );	}
	}

    docRtfWriteDestinationEnd( pCol, sos );
    docRtfWriteNextLine( pCol, sos );

    return;
    }

