/************************************************************************/
/*									*/
/*  Delimit tables and table rectangles.				*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>

#   include	<appDebugon.h>

#   include	"docBuf.h"

/************************************************************************/
/*									*/
/*  Find the range of rows for a 'table'				*/
/*									*/
/************************************************************************/

int docDelimitTable(	BufferItem *		paraBi,
			BufferItem **		pSectBi,
			int *			pCol,
			int *			pRow0,
			int *			pRow,
			int *			pRow1 )
    {
    BufferItem *	rowBi;
    BufferItem *	sectBi;
    int			col;

    if  ( paraBi->biLevel != DOClevPARA		||
	  ! paraBi->biParaInTable		)
	{ /* LLDEB(paraBi->biLevel,paraBi->biParaInTable); */ return -1; }

    rowBi= paraBi->biParent->biParent;
    sectBi= rowBi->biParent;
    col= paraBi->biParent->biNumberInParent;

    *pSectBi= sectBi; *pCol= col;
    *pRow0= rowBi->biRowTableFirst;
    *pRow= rowBi->biNumberInParent;
    *pRow1= rowBi->biRowTablePast- 1;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Determine the rectangular area as it is selected in a table.	*/
/*									*/
/*  This funtion fails if the whole selection is not inside ONE table.	*/
/*									*/
/************************************************************************/

int docGetTableRectangle(	TableRectangle *		tr,
				const DocumentSelection *	ds )
    {
    BufferItem *	sectBi0;
    BufferItem *	sectBi1;

    TableRectangle	trBegin;
    TableRectangle	trEnd;

    if  ( docDelimitTable( ds->dsBegin.dpBi, &sectBi0,
						&(trBegin.trCol0),
						&(trBegin.trRow00),
						&(trBegin.trRow0),
						&(trBegin.trRow11) ) )
	{ /* LDEB(1); */ return -1;	}

    if  ( docDelimitTable( ds->dsEnd.dpBi, &sectBi1,
						&(trEnd.trCol0),
						&(trEnd.trRow00),
						&(trEnd.trRow0),
						&(trEnd.trRow11) ) )
	{ /* LDEB(1); */ return -1;	}

    if  ( sectBi0 != sectBi1 )
	{ /* XXDEB(sectBi0,sectBi1); */ return -1;	}

    if  ( trBegin.trRow00 != trEnd.trRow00	||
	  trBegin.trRow11 != trEnd.trRow11	)
	{ return -1;	}

    if  ( trBegin.trCol0 > trEnd.trCol0 )
	{ return -1;	}
    if  ( trBegin.trRow0 > trEnd.trRow0 )
	{ return -1;	}

    tr->trCol0= trBegin.trCol0;
    tr->trCol1= trEnd.trCol0;
    tr->trCol11= sectBi0->biChildren[trBegin.trRow0]->biChildCount- 1;

    tr->trRow00= trBegin.trRow00;
    tr->trRow0= trBegin.trRow0;
    tr->trRow1= trEnd.trRow0;
    tr->trRow11= trBegin.trRow11;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Get a slice from a table. Either horizontal or vertical or both:	*/
/*  the table as a whole.						*/
/*									*/
/************************************************************************/

int docGetTableSliceSelection(	int *				pIsRowSlice,
				int *				pIsColSlice,
				TableRectangle *		tr,
				const DocumentSelection *	ds )
    {
    int			isRowSlice= 0;
    int			isColSlice= 0;
    TableRectangle	trTry;

    if  ( docGetTableRectangle( &trTry, ds ) )
	{ return 1;	}

    if  ( trTry.trRow0 == trTry.trRow00 && trTry.trRow1 == trTry.trRow11 )
	{ isColSlice= 1;	}
    if  ( trTry.trCol0 == 0 && trTry.trCol1 == trTry.trCol11 )
	{ isRowSlice= 1;	}

    if  ( ! isRowSlice && ! isColSlice )
	{ return 1;	}

    *tr= trTry;
    *pIsRowSlice= isRowSlice;
    *pIsColSlice= isColSlice;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Delimit the tables in a section.					*/
/*									*/
/************************************************************************/

void docSectDelimitTables(	BufferItem *		sectBi )
    {
    int		n;

    if  ( sectBi->biLevel != DOClevSECT )
	{ LDEB(sectBi->biLevel); return;	}

    n= 0;
    while( n < sectBi->biChildCount )
	{
	BufferItem *	rowBiFirst;
	BufferItem *	rowBi;

	int		first;
	int		past;

	rowBiFirst= sectBi->biChildren[n];
	if  ( ! rowBiFirst->biRowHasTableParagraphs )
	    {
	    rowBiFirst->biRowTableFirst= -1;
	    rowBiFirst->biRowTablePast= -1;
	    rowBiFirst->biRowTableFirstIsHeader= 0;

	    n++; continue;
	    }

	first= n;
	past= n+ 1;

	while( past < sectBi->biChildCount )
	    {
	    rowBi= sectBi->biChildren[past];

	    if  ( ! docAlignedColumns( &(rowBiFirst->biRowProperties),
					&(rowBi->biRowProperties) )	)
		{ break;	}

	    if  ( ! docEqualRowBorders( &(rowBiFirst->biRowProperties),
					&(rowBi->biRowProperties) )	)
		{ break;	}

	    past++; 
	    }

	while( n < past )
	    {
	    rowBi= sectBi->biChildren[n];

	    rowBi->biRowTableFirst= first;
	    rowBi->biRowTablePast= past;
	    rowBi->biRowTableFirstIsHeader= rowBiFirst->biRowIsTableHeader;

	    n++;
	    }
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Decide what top border to use for a table cell (If any)		*/
/*									*/
/*  1)  Get the table row above the current row.			*/
/*  2)  If a cell has a bottom boder, and the next one a top border,	*/
/*	MS-Word only draws the bottom border of the topmost cell.	*/
/*  3)  By default use the cells top border.				*/
/*  4)  Use the table top border at the top of a page.			*/
/*									*/
/************************************************************************/

void docGetCellTopBorder(	const BorderProperties **	pBpTop,
				int *				pUseAbove,
				const BufferItem *		rowBi,
				int				col,
				int				atTopOfPage )
    {
    const BufferItem *		sectBi= rowBi->biParent;
    const CellProperties *	cp= &(rowBi->biRowCells[col]);

    const BufferItem *		rowBiAbove= (const BufferItem *)0;

    const CellProperties *	aboveCpR= (const CellProperties *)0;
    const CellProperties *	aboveCpC= (const CellProperties *)0;
    int				useAbove= 0;

    const BorderProperties *	bpTop;

    /*  1  */
    if  ( rowBi->biNumberInParent > rowBi->biRowTableFirst )
	{
	if  ( rowBi->biNumberInParent <= 0 )
	    { LDEB(rowBi->biNumberInParent);	}
	else{ rowBiAbove= sectBi->biChildren[rowBi->biNumberInParent- 1]; }
	}

    if  ( rowBiAbove )
	{ aboveCpR= rowBiAbove->biRowCells;	}

    if  ( rowBiAbove && col < rowBiAbove->biRowCellCount )
	{ aboveCpC= aboveCpR+ col;				}

    /*  2  */
    if  ( aboveCpC && aboveCpC->cpBottomBorder.bpStyle != DOCbsNONE )
	{ useAbove= 1;	}

    /*  3  */
    bpTop= &(cp->cpTopBorder);

    /*  2  */
    if  ( useAbove )
	{ bpTop= &(aboveCpC->cpBottomBorder);	}

    /*  4  */
    if  ( atTopOfPage )
	{
	bpTop= &(rowBi->biRowTopBorder);
	useAbove= 0;
	}

    *pBpTop= bpTop;
    *pUseAbove= useAbove;

    return;
    }

/************************************************************************/
/*									*/
/*  Decide what bottom border to use for a table cell (If any)		*/
/*									*/
/*  1)  Get the table row below the current row.			*/
/*  2)  By default use the cells bottom border.				*/
/*  3)  If a cell has a bottom boder, and the next one a top border,	*/
/*	MS-Word only draws the bottom border of the topmost cell.	*/
/*  4)  Use the table bottom border at the bottom of a page.		*/
/*									*/
/************************************************************************/

void docGetCellBottomBorder(	const BorderProperties **	pBpBottom,
				int *				pUseBelow,
				const BufferItem *		rowBi,
				int				col,
				int				atBottomOfPage )
    {
    const BufferItem *		sectBi= rowBi->biParent;
    const CellProperties *	cp= &(rowBi->biRowCells[col]);

    const BufferItem *		rowBiBelow= (const BufferItem *)0;

    const CellProperties *	belowCpR= (const CellProperties *)0;
    const CellProperties *	belowCpC= (const CellProperties *)0;
    int				useBelow= 0;

    const BorderProperties *	bpBottom;

    /*  1  */
    if  ( rowBi->biNumberInParent < rowBi->biRowTablePast- 1 )
	{ rowBiBelow= sectBi->biChildren[rowBi->biNumberInParent+ 1];	}

    if  ( rowBiBelow )
	{ belowCpR= rowBiBelow->biRowCells;	}

    if  ( rowBiBelow && col < rowBiBelow->biRowCellCount )
	{ belowCpC= belowCpR+ col;				}

    /*  2  */
    bpBottom= &(cp->cpBottomBorder);

    /*  3  */
    if  ( cp->cpBottomBorder.bpStyle == DOCbsNONE	&&
	  belowCpC					&&
	  belowCpC->cpTopBorder.bpStyle != DOCbsNONE	)
	{ useBelow= 1;	}

    /*  4  */
    if  ( atBottomOfPage )
	{
	bpBottom= &(rowBi->biRowBottomBorder);
	useBelow= 0;
	}

    *pBpBottom= bpBottom;
    *pUseBelow= useBelow;

    return;
    }

