/************************************************************************/
/*									*/
/*  Font administration for a document.					*/
/*									*/
/************************************************************************/

#   include	"appUtilConfig.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<string.h>

#   include	<appDebugon.h>

#   include	"utilPropMask.h"
#   include	"docFont.h"
#   include	"utilFontEncoding.h"

/************************************************************************/
/*									*/
/*  Initialise a Font List.						*/
/*  Clean a Font List.							*/
/*  Copy a Font List.							*/
/*									*/
/************************************************************************/

void docInitFontList(	DocumentFontList *	dfl )
    {
    dfl->dflCount= 0;
    dfl->dflFonts= (DocumentFont *)0;
    }

void docCleanFontList(	DocumentFontList *	dfl )
    {
    int		i;

    for ( i = 0; i < dfl->dflCount; i++ )
	{ docCleanFont( dfl->dflFonts+ i );	}

    if  ( dfl->dflFonts )
	{ free( dfl->dflFonts );	}
    }

int docCopyFontList(	DocumentFontList *		to,
			const DocumentFontList *	from )
    {
    int			i;
    DocumentFontList	copy;

    docInitFontList( &copy );

    copy.dflFonts= malloc( from->dflCount* sizeof(DocumentFont) );
    if  ( ! copy.dflFonts )
	{ XDEB(copy.dflFonts); docCleanFontList( &copy ); return -1;	}

    for ( i = 0; i < from->dflCount; i++ )
	{ docInitFont( copy.dflFonts+ i );	}
    copy.dflCount= from->dflCount;

    for ( i = 0; i < from->dflCount; i++ )
	{
	if  ( docCopyFont( copy.dflFonts+ i, from->dflFonts+ i ) )
	    { LDEB(i); docCleanFontList( &copy ); return -1;	}
	}

    docCleanFontList( to );

    *to= copy; return 0;
    }

/************************************************************************/
/*									*/
/*  Initialise a font.							*/
/*  Clean a font.							*/
/*  Copy a font.							*/
/*									*/
/************************************************************************/

void docInitFont(	DocumentFont *	df	)
    {
    df->dfFamilyStyle= (char *)0;
    df->dfName= (char *)0;
    df->dfAltName= (char *)0;
    df->dfDocFamilyNumber= -1;
    df->dfPsFamilyNumber= -1;
    df->dfInstanceCount= 0;
    df->dfInstances= (DocumentFontInstance *)0;
    df->dfCharset= FONTcharsetDEFAULT;
    df->dfEncodingUsed= -1;
    df->dfPitch= FONTpitchDEFAULT;

    df->dfPanose[0]= '\0';
    }

void docCleanFont(	DocumentFont *	df	)
    {
    if  ( df->dfFamilyStyle )
	{ free( df->dfFamilyStyle );	}

    if  ( df->dfName )
	{ free( df->dfName );	}

    if  ( df->dfAltName )
	{ free( df->dfAltName );	}

    if  ( df->dfInstances )
	{ free( df->dfInstances );	}
    }

int docCopyFont(	DocumentFont *		to,
			const DocumentFont *	from )
    {
    DocumentFont	copy;

    copy= *from;

    copy.dfFamilyStyle= (char *)0;
    copy.dfName= (char *)0;
    copy.dfAltName= (char *)0;
    copy.dfInstances= (DocumentFontInstance *)0;

    if  ( from->dfFamilyStyle )
	{
	copy.dfFamilyStyle= strdup( from->dfFamilyStyle );
	if  ( ! copy.dfFamilyStyle )
	    { XDEB(copy.dfFamilyStyle); docCleanFont( &copy ); return -1; }
	}

    if  ( from->dfName )
	{
	copy.dfName= strdup( from->dfName );
	if  ( ! copy.dfName )
	    { XDEB(copy.dfName); docCleanFont( &copy ); return -1; }
	}

    if  ( from->dfAltName )
	{
	copy.dfAltName= strdup( from->dfAltName );
	if  ( ! copy.dfAltName )
	    { XDEB(copy.dfAltName); docCleanFont( &copy ); return -1; }
	}

    if  ( from->dfInstanceCount > 0 )
	{
	int	i;

	copy.dfInstances= (DocumentFontInstance *)
	    malloc( from->dfInstanceCount* sizeof( DocumentFontInstance ) );

	for ( i= 0; i < from->dfInstanceCount; i++ )
	    { copy.dfInstances[i]= from->dfInstances[i];	}
	}

    docCleanFont( to );

    *to= copy; return 0;
    }

/************************************************************************/
/*									*/
/*  Insert a font.							*/
/*									*/
/************************************************************************/

DocumentFont * docInsertFont(	DocumentFontList *	dfl,
				int			n,
				const char *		familyStyle,
				const char *		fontName,
				const char *		fontAltName )
    {
    int			i;
    DocumentFont *	df;

    DocumentFont	dfNew;

    docInitFont( &dfNew );

    if  ( n > dfl->dflCount )
	{
	df= (DocumentFont *)realloc( dfl->dflFonts,
		    ( n + 1 ) * sizeof( DocumentFont ) );
	}
    else{
	df= (DocumentFont *)realloc( dfl->dflFonts,
		    ( dfl->dflCount + 1 ) * sizeof( DocumentFont ) );
	}
    if  ( ! df )
	{ LLDEB(dfl->dflCount,df); docCleanFont( &dfNew ); return df; }
    dfl->dflFonts= df;

    if  ( familyStyle && familyStyle[0] )
	{
	dfNew.dfFamilyStyle= strdup( familyStyle );
	if  ( ! dfNew.dfFamilyStyle )
	    {
	    XDEB(dfNew.dfFamilyStyle);
	    docCleanFont( &dfNew );
	    return (DocumentFont *)0;
	    }
	}

    if  ( fontName && fontName[0] )
	{
	dfNew.dfName= strdup( fontName );
	if  ( ! dfNew.dfName )
	    {
	    XDEB(dfNew.dfName);
	    docCleanFont( &dfNew );
	    return (DocumentFont *)0;
	    }
	}

    if  ( fontAltName && fontAltName[0] )
	{
	dfNew.dfAltName= strdup( fontAltName );
	if  ( ! dfNew.dfAltName )
	    {
	    XDEB(dfNew.dfAltName);
	    docCleanFont( &dfNew );
	    return (DocumentFont *)0;
	    }
	}

    if  ( n == -1 )
	{
	df= dfl->dflFonts;
	for ( n= 0; n < dfl->dflCount; df++, n++ )
	    {
	    if  (  df->dfDocFamilyNumber < 0 )
		{ break;	}
	    }
	}
    else{
	if  ( n < dfl->dflCount && dfl->dflFonts[n].dfDocFamilyNumber >= 0 )
	    { LLDEB(n,dfl->dflFonts[n].dfDocFamilyNumber);	}

	/* No!!
	df= dfl->dflFonts;
	for ( i= dfl->dflCount; i > n; i-- )
	    { df[i]= df[i-1];	}
	*/
	}

    df= dfl->dflFonts+ dfl->dflCount;
    for ( i= dfl->dflCount; i <= n; df++, i++ )
	{ docInitFont( df );	}

    dfNew.dfDocFamilyNumber= n;

    df= dfl->dflFonts+ n;
    docCleanFont( df );

    *df= dfNew;

    if  ( n >= dfl->dflCount )
	{ dfl->dflCount= n+ 1;	}

    return df;
    }

/************************************************************************/
/*									*/
/*  Find or insert a font in a document font list.			*/
/*									*/
/*  In office documents, the same font in PostScript terms can occur	*/
/*  several times in the font list with a different character set and	*/
/*  with a suffix to the family name to differentiate between the	*/
/*  occurences.								*/
/*									*/
/*  NOTE that the name of a font is the key field in this		*/
/*	administration. Character set and family style are dependent	*/
/*	fields.								*/
/*									*/
/************************************************************************/

int docGetFontByName(	DocumentFontList *	dfl,
			const char *		fontFamilyName,
			int			encoding,
			int			encodingIsDefault )
    {
    int			i;
    DocumentFont *	df;

    char *		allocated= (char *)0;
    int			charset= FONTcharsetDEFAULT;

    if  ( encoding >= 0 )
	{
	const FontCharset *	fc= PS_Encodings+ encoding;

	if  ( PS_Encodings[encoding].fcOfficeFontnameSuffix )
	    {
	    allocated= malloc( strlen( fontFamilyName )+ 
				strlen( fc->fcOfficeFontnameSuffix )+ 1 );
	    
	    if  ( ! allocated )
		{ XDEB(allocated); return -1;	}

	    strcpy( allocated, fontFamilyName );
	    strcat( allocated, fc->fcOfficeFontnameSuffix );

	    fontFamilyName= allocated;
	    }

	if  ( ! encodingIsDefault )
	    { charset= fc->fcOfficeCharset;	}
	}

    df= dfl->dflFonts;
    for ( i= 0; i < dfl->dflCount; df++, i++ )
	{
	if  ( df->dfDocFamilyNumber < 0	||
	      ! df->dfName		)
	    { continue;	}

	if  ( ! strcmp( df->dfName, fontFamilyName ) )
	    { break;	}
	}

    if  ( i >= dfl->dflCount )
	{
	const char *	fontFamilyStyle;
	const char *	fontAltName= (const char *)0;

	fontFamilyStyle= utilFontFamilyStyle( fontFamilyName );

	df= docInsertFont( dfl, -1, fontFamilyStyle,
					    fontFamilyName, fontAltName );
	if  ( ! df )
	    { LDEB(dfl->dflCount); return -1;	}

	df->dfCharset= charset;
	}

    if  ( allocated )
	{ free( allocated );	}

    return df->dfDocFamilyNumber;
    }

int docMergeFontIntoFontlist(	DocumentFontList *	dflTo,
				const DocumentFont *	dfFrom )
    {
    int			i;
    DocumentFont *	dfTo;

    dfTo= dflTo->dflFonts;
    for ( i= 0; i < dflTo->dflCount; dfTo++, i++ )
	{
	if  ( dfTo->dfDocFamilyNumber < 0 )
	    { continue;	} 
	if  ( ! dfTo->dfFamilyStyle )
	    { XDEB(dfTo->dfFamilyStyle); continue;	}

	if  ( ! strcmp( dfTo->dfName, dfFrom->dfName )	&&
	      dfTo->dfCharset == dfFrom->dfCharset	)
	    { break;	}
	}

    if  ( i < dflTo->dflCount )
	{ return i;	}

    dfTo= docInsertFont( dflTo, -1,
		dfFrom->dfFamilyStyle, dfFrom->dfName, dfFrom->dfAltName );
    if  ( ! dfTo )
	{ XDEB(dfTo); return -1;	}

    dfTo->dfCharset= dfFrom->dfCharset;

    return dfTo->dfDocFamilyNumber;
    }

