#   include	"bitmapConfig.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<sioStdio.h>
#   include	"bmintern.h"

#   include	<utilPs.h>

#   define	y0	math_y0
#   define	y1	math_y1
#   include	<math.h>
#   undef	y0
#   undef	y1

#   include	<appDebugon.h>

/*
#   define	MAX_BPS		8
#   define	MAX_BPS		12
*/
#   define	MAX_BPS		12

/************************************************************************/
/*									*/
/*  We do our best to print anything.					*/
/*									*/
/************************************************************************/

int bmCanWriteEpsFile(	const BitmapDescription *	bd,
			int				PrivateFormat,
			double				CompressionFactor )
    {
    switch( bd->bdBitsPerSample )
	{
	case 1: case 2: case 4: case 8: case 16:
	    return 0;
	default:
	    /* LDEB(bd->bdBitsPerSample); */ return -1;	
	}
    }

/************************************************************************/
/*									*/
/*  Print Black/White/Color data in hexadecimal format.			*/
/*									*/
/************************************************************************/

static void bmPsWriteShortRow(	SimpleOutputStream *	sos,
				int *			pPos,
				int			pos,
				int			invert,
				int			shortsPerRow,
				const BmUint16 *	from )
    {
    int		col;
    int		val;

#   if MAX_BPS == 8
    const char * const	format= "%02x";
    const int		step= 2;
    const unsigned int	mask= 0xff;
    const unsigned int	shift= 8;
#   endif

#   if MAX_BPS == 12
    const char * const	format= "%03x";
    const int		step= 3;
    const unsigned int	mask= 0xfff;
    const unsigned int	shift= 4;
#   endif

    if  ( invert )
	{
	for ( col= 0; col < shortsPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    val= ~*(from++);
	    val= ( val >> shift ) & mask;

	    sioOutPrintf( sos, format, val );
	    pos += step;
	    }
	}
    else{
	for ( col= 0; col < shortsPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    val= *(from++);
	    val= ( val >> shift ) & mask;

	    sioOutPrintf( sos, format, val );
	    pos += step;
	    }
	}

    *pPos= pos; return;
    }

static void bmPsWriteByteRow0(	SimpleOutputStream *	sos,
				int *			pPos,
				int			pos,
				int			invert,
				int			bytesPerRow,
				const unsigned char *	from )
    {
    int		col;

    if  ( invert )
	{
	for ( col= 0; col < bytesPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    sioOutPrintf( sos, "%02x", ( ~*(from++) ) & 0xff );
	    pos += 2;
	    }
	}
    else{
	for ( col= 0; col < bytesPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    sioOutPrintf( sos, "%02x", *(from++) & 0xff );
	    pos += 2;
	    }
	}

    *pPos= pos; return;
    }

static void bmPsWriteByteRowX(	SimpleOutputStream *	sos,
				int *			pPos,
				int			pos,
				int			invert,
				int			originBit,
				int			rightShift,
				int			bytesPerRow,
				const unsigned char *	from )
    {
    int		col;

    if  ( invert )
	{
	for ( col= 1; col < bytesPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    sioOutPrintf( sos, "%02x",
		( ~( ( from[0] << originBit )	|
		     ( from[1] >> rightShift )	) ) &0xff );

	    from++; pos += 2;
	    }

	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	sioOutPrintf( sos, "%02x", ( ~( from[0] << originBit ) ) &0xff );
	}
    else{
	for ( col= 1; col < bytesPerRow; col++ )
	    {
	    if  ( pos >= 72 )
		{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

	    sioOutPrintf( sos, "%02x",
			    ( from[0] << originBit )	|
			    ( from[1] >> rightShift )	);

	    from++; pos += 2;
	    }

	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	sioOutPrintf( sos, "%02x", from[0] << originBit );
	}

    *pPos= pos; return;
    }

static void bmPsWritePal8Row(	SimpleOutputStream *		sos,
				int *				pPos,
				int				pos,
				const BitmapDescription *	bd,
				int				bytesPerRow,
				const unsigned char *		from )
    {
    int		col;

    for ( col= 0; col < bytesPerRow; col++ )
	{
	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[*from].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[*from].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[*from].rgb8Blue );

	from++; pos += 6;
	}

    *pPos= pos; return;
    }

static void bmPsWritePal4Row0(	SimpleOutputStream *		sos,
				int *				pPos,
				int				pos,
				const BitmapDescription *	bd,
				int				selWidePix,
				const unsigned char *		from )
    {
    int		col;
    int		val;

    for ( col= 0; col+ 1 < selWidePix; col += 2 )
	{
	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	val= ( *from >> 4 ) & 0x0f;

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

	val= *from & 0x0f;

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

	from++; pos += 6;
	}

    if  ( selWidePix % 2 )
	{
	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	val= ( *from >> 4 ) & 0x0f;

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );
	}

    *pPos= pos; return;
    }

static void bmPsWritePal4RowX(	SimpleOutputStream *		sos,
				int *				pPos,
				int				pos,
				const BitmapDescription *	bd,
				int				bytesPerRow,
				const unsigned char *		from )
    {
    int		col;
    int		val;

    if  ( pos >= 72 )
	{ sioOutPutCharacter( '\n', sos ); pos= 0;	}

    val= *from & 0x0f;

    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

    from++; pos += 6;

    for ( col= 1; col < bytesPerRow; col++ )
	{
	if  ( pos >= 72 )
	    { sioOutPutCharacter( '\n', sos ); pos= 0;	}

	val= ( *from >> 4 ) & 0x0f;

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

	val= *from & 0x0f;

	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
	sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

	from++; pos += 6;
	}

    val= ( *from >> 4 ) & 0x0f;

    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Red );
    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Green );
    sioOutPrintf( sos, "%02x", bd->bdRGB8Palette[val].rgb8Blue );

    pos += 6;

    *pPos= pos; return;
    }

int bmPsWriteBitmapData(	    SimpleOutputStream *	sos,
				    int				selX0Pix,
				    int				selY0Pix,
				    int				selWidePix,
				    int				selHighPix,
				    const BitmapDescription *	bd,
				    const unsigned char *	inputBuffer )
    {
    int				row;
    int				pos= 0;

    int				originBit;
    int				originByte;
    int				rightShift;

    int				bytesPerRow;

    const unsigned char *	from;

    if  ( bd->bdBitsPerSample == 16				&&
	  ( bd->bdColorEncoding == BMcoWHITEBLACK	||
	    bd->bdColorEncoding == BMcoBLACKWHITE	||
	    bd->bdColorEncoding == BMcoRGB		)	)
	{
	int		invert= bd->bdColorEncoding == BMcoBLACKWHITE;
	int		shortsPerRow;

	originByte= 2* selX0Pix;
	shortsPerRow= ( selWidePix* bd->bdBitsPerPixel+ 15 )/ 16;

	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWriteShortRow( sos, &pos, pos, invert, shortsPerRow,
						    (const BmUint16 *)from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    originBit=   ( selX0Pix*   bd->bdBitsPerPixel )% 8;
    originByte=  ( selX0Pix*   bd->bdBitsPerPixel )/ 8;
    bytesPerRow= ( selWidePix* bd->bdBitsPerPixel+ 7 )/ 8;
    rightShift= 8- originBit;

    if  ( originBit == 0					&&
	  ( bd->bdColorEncoding == BMcoWHITEBLACK	||
	    bd->bdColorEncoding == BMcoBLACKWHITE	||
	    bd->bdColorEncoding == BMcoRGB		)	)
	{
	int	invert= bd->bdColorEncoding == BMcoBLACKWHITE;

	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWriteByteRow0( sos, &pos, pos, invert, bytesPerRow, from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    if  ( bd->bdColorEncoding == BMcoWHITEBLACK	||
	  bd->bdColorEncoding == BMcoBLACKWHITE )
	{
	int	invert= bd->bdColorEncoding == BMcoBLACKWHITE;

	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWriteByteRowX( sos, &pos, pos, invert,
				originBit, rightShift, bytesPerRow, from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    if  ( bd->bdColorEncoding == BMcoRGB8PALETTE	&&
	  bd->bdBitsPerPixel == 8			)
	{
	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWritePal8Row( sos, &pos, pos, bd, bytesPerRow, from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    if  ( bd->bdColorEncoding == BMcoRGB8PALETTE	&&
	  bd->bdBitsPerPixel == 4			&&
	  originBit == 0				)
	{
	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWritePal4Row0( sos, &pos, pos, bd, selWidePix, from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    if  ( bd->bdColorEncoding == BMcoRGB8PALETTE	&&
	  bd->bdBitsPerPixel == 4			&&
	  originBit == 4				)
	{
	for ( row= 0; row < selHighPix; row++ )
	    {
	    from= inputBuffer+
			( row+ selY0Pix )* bd->bdBytesPerRow+ originByte;

	    bmPsWritePal4RowX( sos, &pos, pos, bd, bytesPerRow, from );
	    }

	sioOutPrintf( sos, "\n" );
	return 0;
	}

    LDEB(bd->bdBitsPerPixel);
    LLDEB(originBit,bd->bdColorEncoding); return -1;
    }

/************************************************************************/
/*									*/
/*  Emit code for the emulation of color images on gray scale printers.	*/
/*									*/
/*  I do not understand this either but it works.			*/
/*									*/
/************************************************************************/

static void bmEpsColorEmulation( SimpleOutputStream *	sos )
    {
    sioOutPutString( "\
	/bwproc {\n\
	    rgbproc\n\
	    dup length 3 idiv string 0 3 0\n\
	    5 -1 roll {\n\
		add 2 1 roll 1 sub dup 0 eq {\n\
		    pop 3 idiv\n\
		    3 -1 roll\n\
		    dup 4 -1 roll\n\
		    dup 3 1 roll\n\
		    5 -1 roll put\n\
		    1 add 3 0\n\
		} { 2 1 roll } ifelse\n\
	    } forall\n\
	    pop pop pop\n\
	} def\n\
	/colorimage where {pop} {\n\
	    /colorimage {pop pop /rgbproc exch def {bwproc} image} bind def\n\
	} ifelse\n", sos );
    }

/************************************************************************/
/*									*/
/*  Return the size of a PostScript string that can hold one row of	*/
/*  data for the bitmap.						*/
/*									*/
/************************************************************************/

int bmPsRowStringSize(		const BitmapDescription *	bd,
				int				pixelsWide )
    {
    switch( bd->bdColorEncoding )
	{
	case	BMcoWHITEBLACK:
	case	BMcoBLACKWHITE:
	case	BMcoRGB:
	    if  ( bd->bdBitsPerSample == 16 )
		{
		int bpp= MAX_BPS* bd->bdSamplesPerPixel;

		return ( pixelsWide* bpp+ 7 )/ 8;
		}
	    else{
		return ( pixelsWide* bd->bdBitsPerPixel+ 7 )/ 8;
		}

	case	BMcoRGB8PALETTE:
	    return ( pixelsWide* 24+ 7 )/ 8;

	default:
	    LDEB(bd->bdColorEncoding); return -1;
	}
    }

/************************************************************************/
/*									*/
/*  Print one Encapsulated postsctipt page.				*/
/*									*/
/************************************************************************/

int bmPsPrintBitmap(	SimpleOutputStream *		sos,
			int				level,
			double				xscale,
			double				yscale,
			int				ox,
			int				oy,
			int				selX0Pix,
			int				selY0Pix,
			int				selWidePix,
			int				selHighPix,
			const BitmapDescription *	bd,
			const unsigned char *		buffer	)
    {
    switch( bd->bdUnit )
	{
	case BMunM:
	    xscale= ( xscale* POINTS_PER_M* selWidePix )/ bd->bdXResolution;
	    yscale= ( yscale* POINTS_PER_M* selHighPix )/ bd->bdYResolution;
	    break;

	case BMunINCH:
	    xscale= ( xscale* 72.0* selWidePix )/ bd->bdXResolution;
	    yscale= ( yscale* 72.0* selHighPix )/ bd->bdYResolution;
	    break;

	case BMunPOINT:
	    xscale= ( xscale* 1.0* selWidePix )/ bd->bdXResolution;
	    yscale= ( yscale* 1.0* selHighPix )/ bd->bdYResolution;
	    break;

	case BMunPIXEL:
	    xscale= ( xscale* 1.0* selWidePix );
	    yscale= ( yscale* 1.0* selHighPix );
	    break;

	default:
	    LDEB(bd->bdUnit);
	    xscale= ( xscale* 1.0* selWidePix );
	    yscale= ( yscale* 1.0* selHighPix );
	    break;
	}

    return bmPsPrintBitmapImage( sos, level, xscale, yscale, ox, oy,
				    selX0Pix, selY0Pix, selWidePix, selHighPix,
				    bd, buffer );
    }

void bmPsWriteImageInstructions(
				SimpleOutputStream *		sos,
				const BitmapDescription *	bd,
				int				selWidePix,
				int				selHighPix,
				const char *			proc )
    {
    int		bitsPerSample= bd->bdBitsPerSample;

    if  ( bitsPerSample > MAX_BPS )
	{ bitsPerSample= MAX_BPS;	}

    switch( bd->bdColorEncoding )
	{
	case	BMcoWHITEBLACK:
	case	BMcoBLACKWHITE:
	    if  ( bd->bdBitsPerSample == 1 )
		{ sioOutPrintf( sos, "0 setgray\n" );	}

	    sioOutPrintf( sos, "%u %u", selWidePix, selHighPix );

	    if  ( bd->bdBitsPerSample == 1 )
		{ sioOutPrintf( sos, " false\n" );			}
	    else{ sioOutPrintf( sos, " %d\n", bitsPerSample );	}

	    sioOutPrintf( sos, "[ %u 0 0 -%u 0 %u ]\n",
				    selWidePix, selHighPix, selHighPix );
	    sioOutPrintf( sos, "{ %s } bind\n", proc );

	    if  ( bd->bdBitsPerSample == 1 )
		{ sioOutPrintf( sos, "imagemask\n" );		}
	    else{ sioOutPrintf( sos, "image\n" );		}

	    break;
	
	case	BMcoRGB:
	case	BMcoRGB8PALETTE:
	    bmEpsColorEmulation( sos );

	    sioOutPrintf( sos, "%u %u %d\n",
			    selWidePix, selHighPix, bitsPerSample );

	    sioOutPrintf( sos, "[ %u 0 0 -%u 0 %u ]\n",
			    selWidePix, selHighPix, selHighPix );
	    sioOutPrintf( sos, "{ %s } bind\n", proc );
	    sioOutPrintf( sos, "false %d colorimage\n", 3 );

	    break;

	default:
	    LDEB(bd->bdColorEncoding);
	}

    return;
    }

int bmPsPrintBitmapImage(	SimpleOutputStream *		sos,
				int				level,
				double				xscale,
				double				yscale,
				int				ox,
				int				oy,
				int				selX0Pix,
				int				selY0Pix,
				int				selWidePix,
				int				selHighPix,
				const BitmapDescription *	bd,
				const unsigned char *		buffer	)
    {
    int		bytesPerRow;

    sioOutPrintf( sos, "gsave 10 dict begin %% {> bitmap image\n" );

    if  ( ox != 0 || oy != 0 || xscale != 1.0 || yscale != 1.0 )
	{
	if  ( ox != 0 || oy != 0 )
	    { sioOutPrintf( sos, "%d %d translate ", ox, oy );	}
	if  ( xscale != 1.0 || yscale != 1.0 )
	    { sioOutPrintf( sos, "%f %f scale ", xscale, yscale ); }

	sioOutPrintf( sos, "\n" );
	}

    bytesPerRow= bmPsRowStringSize( bd, selWidePix );
    if  ( bytesPerRow < 0 )
	{ LDEB(bytesPerRow); return -1;	}

    sioOutPrintf( sos, "/line %d string def\n", bytesPerRow );

    bmPsWriteImageInstructions( sos, bd, selWidePix, selHighPix,
					"currentfile line readhexstring pop" );

    bmPsWriteBitmapData( sos, selX0Pix, selY0Pix, selWidePix, selHighPix,
								bd, buffer );

    sioOutPrintf( sos, "end grestore %% }< bitmap image\n");

    return 0;
    }

/************************************************************************/
/*									*/
/*  Calculate the page state parameters.				*/
/*									*/
/************************************************************************/

static void bmPsPageExtent(	const BitmapDescription *	bd,
				float *				pPointsWide,
				float *				pPointsHigh,
				int				pixelsWide,
				int				pixelsHigh )
    {
    double	xResolution;
    double	yResolution;

    switch( bd->bdUnit )
	{
	case BMunM:
	    xResolution= bd->bdXResolution/ POINTS_PER_M;
	    yResolution= bd->bdYResolution/ POINTS_PER_M;
	    break;

	case BMunINCH:
	    xResolution= bd->bdXResolution/ 72.0;
	    yResolution= bd->bdYResolution/ 72.0;
	    break;

	case BMunPOINT:
	    xResolution= bd->bdXResolution;
	    yResolution= bd->bdYResolution;
	    break;

	case BMunPIXEL:
	    xResolution= 1.0;
	    yResolution= 1.0;
	    break;

	default:
	    LDEB(bd->bdUnit); return;
	}

    *pPointsWide= pixelsWide/ xResolution;
    *pPointsHigh= pixelsHigh/ yResolution;

    return;
    }

/************************************************************************/
/*									*/
/*  Save a bitmap image to Encapsulated PostScript (EPS) Format.	*/
/*									*/
/*  1)  Only 1,2,4,8 bits per sample images are supported.		*/
/*  2)  Open the output file.						*/
/*  3)  Determine the bounding box of the image.			*/
/*  4)  Issue the EPS header comments.					*/
/*  5)  Issue the Image contents.					*/
/*  6)  Include a 'showpage' command.					*/
/*  7)  Finish with a general header.					*/
/*									*/
/************************************************************************/

int bmWriteEpsFile(	const char *			filename,
			const unsigned char *		buffer,
			const BitmapDescription *	bd,
			int				privateFormat,
			double				compressionFactor )
    {
    SimpleOutputStream *	sos;
    float			pointsWide;
    float			pointsHigh;

    /*  1  */
    switch( bd->bdBitsPerSample )
	{
	case 1:
	case 2:
	case 4:
	case 8:
	    break;
	default:
	    LDEB(bd->bdBitsPerSample); return -1;	
	}

    /*  2  */
    sos= sioOutStdioOpen( filename );
    if  ( ! sos )
	{ SXDEB(filename,sos); return -1;	}

    /*  3  */
    bmPsPageExtent( bd, &pointsWide, &pointsHigh,
					bd->bdPixelsWide, bd->bdPixelsHigh );

    /*  4  */
    appPsWriteEpsHeader( sos, 
		    "Scan, Mark de Does, mark@mdedoes.com",
		    filename, 
		    (int)ceil( pointsWide ), (int)ceil( pointsHigh )	);

    /*  5  */
    if  ( bmPsPrintBitmap( sos, privateFormat, 1.0, 1.0, 0, 0,
				0, 0, bd->bdPixelsWide, bd->bdPixelsHigh,
				bd, buffer ) )
	{ LDEB(1); sioOutClose( sos ); return -1;	}

    /*  6  */
    sioOutPrintf( sos, "showpage\n");

    /*  7  */
    sioOutPrintf( sos, "%%%%Trailer\n" );
    sioOutPrintf( sos, "%%%%EOF\n" );

    sioOutClose( sos ); return 0;
    }
